# cython: language_level=3

"""Python code for writing AVRO files"""

# This code is a modified version of the code at
# http://svn.apache.org/viewvc/avro/trunk/lang/py/src/avro/ which is under
# Apache 2.0 license (http://www.apache.org/licenses/LICENSE-2.0)

from cpython cimport array
import array
import json
from binascii import crc32
from os import urandom
import bz2
import lzma
import zlib
import sys

from fastavro import const
from ._logical_writers import LOGICAL_WRITERS
from ._validation import _validate
from ._read import HEADER_SCHEMA, SYNC_SIZE, MAGIC, reader
from ._schema import extract_record_type, extract_logical_type, parse_schema
from ._write_common import _is_appendable

CYTHON_MODULE = 1  # Tests check this to confirm whether using the Cython code.

ctypedef int int32
ctypedef unsigned int uint32
ctypedef unsigned long long ulong64
ctypedef long long long64

cdef long64 MCS_PER_SECOND = const.MCS_PER_SECOND
cdef long64 MCS_PER_MINUTE = const.MCS_PER_MINUTE
cdef long64 MCS_PER_HOUR = const.MCS_PER_HOUR

cdef long64 MLS_PER_SECOND = const.MLS_PER_SECOND
cdef long64 MLS_PER_MINUTE = const.MLS_PER_MINUTE
cdef long64 MLS_PER_HOUR = const.MLS_PER_HOUR


cdef inline write_null(object fo, datum):
    """null is written as zero bytes"""
    pass


cdef inline write_boolean(bytearray fo, bint datum):
    """A boolean is written as a single byte whose value is either 0 (false) or
    1 (true)."""
    cdef unsigned char ch_temp[1]
    ch_temp[0] = 1 if datum else 0
    fo += ch_temp[:1]


cdef inline write_int(bytearray fo, long64 datum):
    """int and long values are written using variable-length, zig-zag coding.
    """
    cdef ulong64 n
    cdef unsigned char ch_temp[1]
    n = (datum << 1) ^ (datum >> 63)
    while (n & ~0x7F) != 0:
        ch_temp[0] = (n & 0x7f) | 0x80
        fo += ch_temp[:1]
        n >>= 7
    ch_temp[0] = n
    fo += ch_temp[:1]


cdef inline write_long(bytearray fo, datum):
    write_int(fo, datum)


cdef union float_uint32:
    float f
    uint32 n


cdef inline write_float(bytearray fo, float datum):
    """A float is written as 4 bytes.  The float is converted into a 32-bit
    integer using a method equivalent to Java's floatToIntBits and then encoded
    in little-endian format."""
    cdef float_uint32 fi
    cdef unsigned char ch_temp[4]

    fi.f = datum
    ch_temp[0] = fi.n & 0xff
    ch_temp[1] = (fi.n >> 8) & 0xff
    ch_temp[2] = (fi.n >> 16) & 0xff
    ch_temp[3] = (fi.n >> 24) & 0xff

    fo += ch_temp[:4]


cdef union double_ulong64:
    double d
    ulong64 n


cdef inline write_double(bytearray fo, double datum):
    """A double is written as 8 bytes.  The double is converted into a 64-bit
    integer using a method equivalent to Java's doubleToLongBits and then
    encoded in little-endian format.  """
    cdef double_ulong64 fi
    cdef unsigned char ch_temp[8]

    fi.d = datum
    ch_temp[0] = fi.n & 0xff
    ch_temp[1] = (fi.n >> 8) & 0xff
    ch_temp[2] = (fi.n >> 16) & 0xff
    ch_temp[3] = (fi.n >> 24) & 0xff
    ch_temp[4] = (fi.n >> 32) & 0xff
    ch_temp[5] = (fi.n >> 40) & 0xff
    ch_temp[6] = (fi.n >> 48) & 0xff
    ch_temp[7] = (fi.n >> 56) & 0xff

    fo += ch_temp[:8]


cdef inline write_bytes(bytearray fo, const unsigned char[:] datum):
    """Bytes are encoded as a long followed by that many bytes of data."""
    write_long(fo, len(datum))
    fo += datum


cdef inline write_utf8(bytearray fo, datum):
    """A string is encoded as a long followed by that many bytes of UTF-8
    encoded character data."""
    b_datum = datum.encode()
    write_long(fo, len(b_datum))
    fo += b_datum


cdef inline write_crc32(bytearray fo, bytes bytes):
    """A 4-byte, big-endian CRC32 checksum"""
    cdef unsigned char ch_temp[4]
    cdef uint32 data = crc32(bytes) & 0xFFFFFFFF

    ch_temp[0] = (data >> 24) & 0xff
    ch_temp[1] = (data >> 16) & 0xff
    ch_temp[2] = (data >> 8) & 0xff
    ch_temp[3] = data & 0xff
    fo += ch_temp[:4]


cdef inline write_fixed(bytearray fo, object datum, dict named_schemas):
    """Fixed instances are encoded using the number of bytes declared in the
    schema."""
    fo += datum


cdef inline write_enum(bytearray fo, datum, schema, dict named_schemas):
    """An enum is encoded by a int, representing the zero-based position of
    the symbol in the schema."""
    index = schema["symbols"].index(datum)
    write_int(fo, index)


cdef write_array(bytearray fo, list datum, schema, dict named_schemas, fname):
    """Arrays are encoded as a series of blocks.

    Each block consists of a long count value, followed by that many array
    items.  A block with count zero indicates the end of the array.  Each item
    is encoded per the array's item schema.

    If a block's count is negative, then the count is followed immediately by a
    long block size, indicating the number of bytes in the block.  The actual
    count in this case is the absolute value of the count written.  """

    if len(datum) > 0:
        write_long(fo, len(datum))
        dtype = schema["items"]
        for item in datum:
            write_data(fo, item, dtype, named_schemas, fname)
    write_long(fo, 0)


cdef write_python_array(bytearray fo,
                        array.array datum,
                        schema,
                        named_schemas,
                        fname):
    """Array specialization for python arrays."""
    if len(datum) > 0:
        write_long(fo, len(datum))
        record_type = extract_record_type(schema["items"])
        if record_type in ("int", "long"):
            for item in datum:
                write_int(fo, item)
        elif record_type == "float":
            for item in datum:
                write_float(fo, item)
        elif record_type == "double":
            for item in datum:
                write_double(fo, item)
        else:
            dtype = schema["items"]
            for item in datum:
                write_data(fo, item, dtype, named_schemas, fname)
    write_long(fo, 0)


cdef write_map(bytearray fo, object datum, dict schema, dict named_schemas, fname):
    """Maps are encoded as a series of blocks.

    Each block consists of a long count value, followed by that many key/value
    pairs.  A block with count zero indicates the end of the map.  Each item is
    encoded per the map's value schema.

    If a block's count is negative, then the count is followed immediately by a
    long block size, indicating the number of bytes in the block. The actual
    count in this case is the absolute value of the count written."""
    cdef dict d_datum
    try:
        d_datum = <dict?>(datum)
    except TypeError:
        # Slower, general-purpose code where datum is something besides a dict,
        # e.g. a collections.OrderedDict or collections.defaultdict.
        if len(datum) > 0:
            write_long(fo, len(datum))
            vtype = schema["values"]
            for key, val in datum.items():
                write_utf8(fo, key)
                write_data(fo, val, vtype, named_schemas, fname)
        write_long(fo, 0)
    else:
        # Faster, special-purpose code where datum is a Python dict.
        if len(d_datum) > 0:
            write_long(fo, len(d_datum))
            vtype = schema["values"]
            for key, val in d_datum.items():
                write_utf8(fo, key)
                write_data(fo, val, vtype, named_schemas, fname)
        write_long(fo, 0)


cdef write_union(bytearray fo, datum, schema, dict named_schemas, fname):
    """A union is encoded by first writing a long value indicating the
    zero-based position within the union of the schema of its value. The value
    is then encoded per the indicated schema within the union."""

    cdef int32 best_match_index
    cdef int32 most_fields
    cdef int32 index
    cdef int32 fields
    cdef str schema_name
    best_match_index = -1
    if isinstance(datum, tuple):
        (name, datum) = datum
        for index, candidate in enumerate(schema):
            if extract_record_type(candidate) == "record":
                schema_name = candidate["name"]
            else:
                schema_name = candidate
            if name == schema_name:
                best_match_index = index
                break

        if best_match_index == -1:
            field = f"on field {fname}" if fname else ""
            msg = (
                f"provided union type name {name} not found in schema "
                + f"{schema} {field}"
            )
            raise ValueError(msg)
        index = best_match_index
    else:
        pytype = type(datum)
        most_fields = -1
        for index, candidate in enumerate(schema):
            if _validate(datum, candidate, named_schemas, raise_errors=False):
                if extract_record_type(candidate) == "record":
                    candidate_fields = set(
                        f["name"] for f in candidate["fields"]
                    )
                    datum_fields = set(datum)
                    fields = len(candidate_fields.intersection(datum_fields))
                    if fields > most_fields:
                        best_match_index = index
                        most_fields = fields
                else:
                    best_match_index = index
                    break
        if best_match_index == -1:
            field = f"on field {fname}" if fname else ""
            raise ValueError(
                f"{repr(datum)} (type {pytype}) do not match {schema} {field}"
            )
        index = best_match_index

    # write data
    write_long(fo, index)
    write_data(fo, datum, schema[index], named_schemas, fname)


cdef write_record(bytearray fo, object datum, dict schema, dict named_schemas):
    """A record is encoded by encoding the values of its fields in the order
    that they are declared. In other words, a record is encoded as just the
    concatenation of the encodings of its fields.  Field values are encoded per
    their schema."""
    cdef list fields
    cdef dict field
    cdef dict d_datum
    fields = schema["fields"]
    try:
        d_datum = <dict?>(datum)
    except TypeError:
        # Slower, general-purpose code where datum is something besides a dict,
        # e.g. a collections.OrderedDict or collections.defaultdict.
        for field in fields:
            name = field["name"]
            if name not in datum and "default" not in field and "null" not in field["type"]:
                raise ValueError(f"no value and no default for {name}")
            datum_value = datum.get(name, field.get("default"))
            write_data(fo, datum_value, field["type"], named_schemas, name)
    else:
        # Faster, special-purpose code where datum is a Python dict.
        for field in fields:
            name = field["name"]
            if name not in d_datum and "default" not in field and \
                    "null" not in field["type"]:
                raise ValueError(f"no value and no default for {name}")
            d_datum_value = datum.get(name, field.get("default"))
            write_data(fo, d_datum_value, field["type"], named_schemas, name)


cpdef write_data(bytearray fo, datum, schema, dict named_schemas, fname):
    """Write a datum of data to output stream.

    Paramaters
    ----------
    fo: file like
        Output file
    datum: object
        Data to write
    schema: dict
        Schema to use
    """
    cdef str logical_type = None
    if isinstance(schema, dict):
        logical_type = extract_logical_type(schema)
        if logical_type:
            prepare = LOGICAL_WRITERS.get(logical_type)
            if prepare:
                datum = prepare(datum, schema)

    record_type = extract_record_type(schema)
    try:
        if record_type == "null":
            return write_null(fo, datum)
        elif record_type == "string":
            return write_utf8(fo, datum)
        elif record_type == "int" or record_type == "long":
            return write_long(fo, datum)
        elif record_type == "float":
            return write_float(fo, datum)
        elif record_type == "double":
            return write_double(fo, datum)
        elif record_type == "boolean":
            return write_boolean(fo, datum)
        elif record_type == "bytes":
            return write_bytes(fo, datum)
        elif record_type == "fixed":
            return write_fixed(fo, datum, named_schemas)
        elif record_type == "enum":
            return write_enum(fo, datum, schema, named_schemas)
        elif record_type == "array":
            if isinstance(datum, array.array):
                return write_python_array(
                    fo, datum, schema, named_schemas, fname
                )
            elif not isinstance(datum, list):
                datum = list(datum)
            return write_array(fo, datum, schema, named_schemas, fname)
        elif record_type == "map":
            return write_map(fo, datum, schema, named_schemas, fname)
        elif record_type == "union" or record_type == "error_union":
            return write_union(fo, datum, schema, named_schemas, fname)
        elif record_type == "record" or record_type == "error":
            return write_record(fo, datum, schema, named_schemas)
        else:
            return write_data(
                fo, datum, named_schemas[record_type], named_schemas, fname
            )
    except TypeError as ex:
        if fname:
            msg = f"{ex} on field {fname}"
            # TODO: Can we use `raise TypeError(msg) from ex` here?
            raise TypeError(msg).with_traceback(sys.exc_info()[2])
        raise


cpdef write_header(bytearray fo, dict metadata, bytes sync_marker):
    header = {
        "magic": MAGIC,
        "meta": {key: value.encode() for key, value in metadata.items()},
        "sync": sync_marker
    }
    write_data(fo, header, HEADER_SCHEMA, {}, "")


cpdef null_write_block(object fo, bytes block_bytes, compression_level):
    """Write block in "null" codec."""
    cdef bytearray tmp = bytearray()
    write_long(tmp, len(block_bytes))
    fo.write(tmp)
    fo.write(block_bytes)


cpdef deflate_write_block(object fo, bytes block_bytes, compression_level):
    """Write block in "deflate" codec."""
    cdef bytearray tmp = bytearray()
    # The first two characters and last character are zlib
    # wrappers around deflate data.
    if compression_level is not None:
        data = zlib.compress(block_bytes, compression_level)[2:-1]
    else:
        data = zlib.compress(block_bytes)[2:-1]

    write_long(tmp, len(data))
    fo.write(tmp)
    fo.write(data)


cpdef bzip2_write_block(object fo, bytes block_bytes, compression_level):
    """Write block in "bzip2" codec."""
    cdef bytearray tmp = bytearray()
    data = bz2.compress(block_bytes)
    write_long(tmp, len(data))
    fo.write(tmp)
    fo.write(data)


cpdef xz_write_block(object fo, bytes block_bytes, compression_level):
    """Write block in "xz" codec."""
    cdef bytearray tmp = bytearray()
    data = lzma.compress(block_bytes)
    write_long(tmp, len(data))
    fo.write(tmp)
    fo.write(data)


BLOCK_WRITERS = {
    "null": null_write_block,
    "deflate": deflate_write_block,
    "bzip2": bzip2_write_block,
    "xz": xz_write_block,
}


def _missing_dependency(codec, library):
    def missing(fo, block_bytes, compression_level):
        raise ValueError(
            f"{codec} codec is supported but you need to install {library}"
        )
    return missing


try:
    import snappy
except ImportError:
    BLOCK_WRITERS["snappy"] = _missing_dependency("snappy", "python-snappy")


cpdef snappy_write_block(object fo, bytes block_bytes, compression_level):
    """Write block in "snappy" codec."""
    cdef bytearray tmp = bytearray()
    data = snappy.compress(block_bytes)

    write_long(tmp, len(data) + 4)  # for CRC
    fo.write(tmp)
    fo.write(data)
    tmp[:] = b""
    write_crc32(tmp, block_bytes)
    fo.write(tmp)


if BLOCK_WRITERS.get("snappy") is None:
    BLOCK_WRITERS["snappy"] = snappy_write_block


try:
    import zstandard as zstd
except ImportError:
    BLOCK_WRITERS["zstandard"] = _missing_dependency("zstandard", "zstandard")


cpdef zstandard_write_block(object fo, bytes block_bytes, compression_level):
    """Write block in "zstandard" codec."""
    cdef bytearray tmp = bytearray()
    data = zstd.ZstdCompressor().compress(block_bytes)
    write_long(tmp, len(data))
    fo.write(tmp)
    fo.write(data)


if BLOCK_WRITERS.get("zstandard") is None:
    BLOCK_WRITERS["zstandard"] = zstandard_write_block


try:
    import lz4.block
except ImportError:
    BLOCK_WRITERS["lz4"] = _missing_dependency("lz4", "lz4")


cpdef lz4_write_block(object fo, bytes block_bytes, compression_level):
    """Write block in "lz4" codec."""
    cdef bytearray tmp = bytearray()
    data = lz4.block.compress(block_bytes)
    write_long(tmp, len(data))
    fo.write(tmp)
    fo.write(data)


if BLOCK_WRITERS.get("lz4") is None:
    BLOCK_WRITERS["lz4"] = lz4_write_block


cdef class MemoryIO:
    cdef bytearray value

    def __init__(self):
        self.value = bytearray()

    cpdef int32 tell(self):
        return len(self.value)

    cpdef bytes getvalue(self):
        return bytes(self.value)

    cpdef clear(self):
        self.value[:] = b""


cdef class Writer:
    cdef public object fo
    cdef public object schema
    cdef public object validate_fn
    cdef public object sync_marker
    cdef public MemoryIO io
    cdef public int32 block_count
    cdef public object metadata
    cdef public long64 sync_interval
    cdef public object block_writer
    cdef public object compression_level
    cdef public dict _named_schemas

    def __init__(self,
                 fo,
                 schema,
                 codec="null",
                 sync_interval=1000 * SYNC_SIZE,
                 metadata=None,
                 validator=None,
                 sync_marker=None,
                 compression_level=None):
        cdef bytearray tmp = bytearray()

        self.fo = fo
        self._named_schemas = {}
        self.schema = parse_schema(schema, _named_schemas=self._named_schemas)
        self.validate_fn = _validate if validator is True else validator
        self.io = MemoryIO()
        self.block_count = 0
        self.sync_interval = sync_interval
        self.compression_level = compression_level

        if _is_appendable(self.fo):
            # Seed to the beginning to read the header
            self.fo.seek(0)
            avro_reader = reader(self.fo)
            header = avro_reader._header

            file_writer_schema = parse_schema(avro_reader.writer_schema)
            if self.schema != file_writer_schema:
                raise ValueError(
                    f"Provided schema {self.schema} does not match file "
                    + f"writer_schema {file_writer_schema}"
                )
            codec = avro_reader.metadata.get("avro.codec", "null")

            self.sync_marker = header["sync"]

            # Seek to the end of the file
            self.fo.seek(0, 2)

            self.block_writer = BLOCK_WRITERS[codec]
        else:
            self.sync_marker = sync_marker or urandom(SYNC_SIZE)

            self.metadata = metadata or {}
            self.metadata["avro.codec"] = codec

            if isinstance(schema, dict):
                schema = {
                    key: value
                    for key, value in schema.items()
                    if key not in ("__fastavro_parsed", "__named_schemas")
                }
            elif isinstance(schema, list):
                schemas = []
                for s in schema:
                    if isinstance(s, dict):
                        schemas.append(
                            {
                                key: value
                                for key, value in s.items()
                                if key not in (
                                    "__fastavro_parsed", "__named_schemas"
                                )
                            }
                        )
                    else:
                        schemas.append(s)
                schema = schemas

            self.metadata["avro.schema"] = json.dumps(schema)

            try:
                self.block_writer = BLOCK_WRITERS[codec]
            except KeyError:
                raise ValueError(f"unrecognized codec: {codec}")

            write_header(tmp, self.metadata, self.sync_marker)
            self.fo.write(tmp)

    def dump(self):
        cdef bytearray tmp = bytearray()
        write_long(tmp, self.block_count)
        self.fo.write(tmp)
        self.block_writer(self.fo, self.io.getvalue(), self.compression_level)
        self.fo.write(self.sync_marker)
        self.io.clear()
        self.block_count = 0

    def write(self, record):
        if self.validate_fn:
            self.validate_fn(record, self.schema, self._named_schemas)
        write_data(self.io.value, record, self.schema, self._named_schemas, "")
        self.block_count += 1
        if self.io.tell() >= self.sync_interval:
            self.dump()

    def write_block(self, block):
        # Clear existing block if there are any records pending
        if self.io.tell() or self.block_count > 0:
            self.dump()
        cdef bytearray tmp = bytearray()
        write_long(tmp, block.num_records)
        self.fo.write(tmp)
        self.block_writer(self.fo, block.bytes_.getvalue(), self.compression_level)
        self.fo.write(self.sync_marker)

    def flush(self):
        if self.io.tell() or self.block_count > 0:
            self.dump()
        self.fo.flush()


def writer(fo,
           schema,
           records,
           codec="null",
           sync_interval=1000 * SYNC_SIZE,
           metadata=None,
           validator=None,
           sync_marker=None,
           codec_compression_level=None):
    # Sanity check that records is not a single dictionary (as that is a common
    # mistake and the exception that gets raised is not helpful)
    if isinstance(records, dict):
        raise ValueError('"records" argument should be an iterable, not dict')

    output = Writer(
        fo,
        schema,
        codec,
        sync_interval,
        metadata,
        validator,
        sync_marker,
        codec_compression_level,
    )

    for record in records:
        output.write(record)
    output.flush()


def schemaless_writer(fo, schema, record):
    cdef bytearray tmp = bytearray()
    named_schemas = {}
    schema = parse_schema(schema, _named_schemas=named_schemas)
    write_data(tmp, record, schema, named_schemas, "")
    fo.write(tmp)



from typing import Dict, Union, List

SchemaType = Union[str, List, Dict]

FINGERPRINT_ALGORITHMS: str

def load_schema(schema_path: str) -> Dict: ...
def parse_schema(
    schema: SchemaType, expand: bool, _write_hint: bool, _force: bool
) -> SchemaType: ...
def fullname(schema: Dict) -> str: ...
def expand_schema(schema: Dict) -> Dict: ...
def load_schema_ordered(ordered_schemas: List[str]) -> Dict: ...
def to_parsing_canonical_form(schema: SchemaType) -> str: ...
def fingerprint(canonical_form: str, algorithm: str) -> str: ...



from typing import Callable, Dict, IO, Iterable, Optional, Union
from .io.binary_encoder import BinaryEncoder
from .types import AvroMessage

def writer(
    fo: IO,
    schema: Dict,
    records: Iterable,
    codec: str,
    sync_interval: int,
    metadata: Optional[Dict],
    validator: Union[Callable, bool, None],
    sync_marker: Optional[bytes],
    codec_compression_level: Optional[int],
) -> None: ...

class GenericWriter:
    schema: Dict
    validate_fn: Callable
    metadata: Dict

class Writer(GenericWriter):
    encoder: BinaryEncoder
    io: BinaryEncoder
    block_count: int
    sync_interval: int
    compression_level: Optional[int]
    block_writer: Callable
    def __init__(
        self,
        fo: IO,
        records: Iterable,
        codec: str,
        sync_interval: int,
        metadata: Optional[Dict],
        validator: Union[Callable, bool, None],
        sync_marker: Optional[bytes],
        codec_compression_level: Optional[int],
    ): ...
    def dump(self) -> None: ...
    def write(self, record: AvroMessage) -> None: ...
    def write_block(self, block) -> None: ...
    def flush(self) -> None: ...

def schemaless_writer(fo: IO, schema: Dict, record: Dict) -> None: ...



from typing import Any, Dict, Iterable, Optional

def validate(
    datum: Any, schema: Dict, field: Optional[str], raise_errors: bool
) -> bool: ...
def validate_many(
    records: Iterable[Dict], schema: Dict, raise_errors: bool
) -> bool: ...



from fastavro import writer as fastavro_writer
from fastavro.read import SchemaResolutionError
import fastavro

import pytest

from io import BytesIO

schema_dict_a = {
    "namespace": "example.avro2",
    "type": "record",
    "name": "evtest",
    "fields": [{"name": "a", "type": "int"}],
}

record_a = {"a": 123}

schema_dict_a_b = {
    "namespace": "example.avro2",
    "type": "record",
    "name": "evtest",
    "fields": [
        {"name": "a", "type": "int"},
        {"name": "b", "type": ["null", "int"], "default": None},
    ],
}

record_a_b = {"a": 234, "b": 345}

schema_dict_a_c = {
    "namespace": "example.avro2",
    "type": "record",
    "name": "evtest",
    "fields": [{"name": "a", "type": "int"}, {"name": "c", "type": ["null", "int"]}],
}


def avro_to_bytes_with_schema(avro_schema, avro_dict):
    with BytesIO() as bytes_io:
        fastavro_writer(bytes_io, avro_schema, [avro_dict])
        return bytes_io.getvalue()


def bytes_with_schema_to_avro(avro_read_schema, binary):
    with BytesIO(binary) as bytes_io:
        reader = fastavro.reader(bytes_io, avro_read_schema)
        return next(reader)


def test_evolution_drop_field():
    record_bytes_a_b = avro_to_bytes_with_schema(schema_dict_a_b, record_a_b)
    record_a = bytes_with_schema_to_avro(schema_dict_a, record_bytes_a_b)
    assert "b" not in record_a


def test_evolution_add_field_with_default():
    record_bytes_a = avro_to_bytes_with_schema(schema_dict_a, record_a)
    record_b = bytes_with_schema_to_avro(schema_dict_a_b, record_bytes_a)
    assert "b" in record_b
    assert record_b.get("b") is None


def test_evolution_add_field_without_default():
    with pytest.raises(SchemaResolutionError):
        record_bytes_a = avro_to_bytes_with_schema(schema_dict_a, record_a)
        bytes_with_schema_to_avro(schema_dict_a_c, record_bytes_a)


def test_enum_evolution_no_default_failure():
    original_schema = {
        "type": "enum",
        "name": "test",
        "symbols": ["FOO", "BAR"],
    }

    new_schema = {
        "type": "enum",
        "name": "test",
        "symbols": ["BAZ", "BAR"],
    }

    original_records = ["FOO"]

    bio = BytesIO()
    fastavro.writer(bio, original_schema, original_records)
    bio.seek(0)

    with pytest.raises(fastavro.read.SchemaResolutionError):
        list(fastavro.reader(bio, new_schema))


def test_enum_evolution_using_default():
    original_schema = {
        "type": "enum",
        "name": "test",
        "symbols": ["A", "B"],
    }

    new_schema = {
        "type": "enum",
        "name": "test",
        "symbols": ["C", "D"],
        "default": "C",
    }

    original_records = ["A"]

    bio = BytesIO()
    fastavro.writer(bio, original_schema, original_records)
    bio.seek(0)

    new_records = list(fastavro.reader(bio, new_schema))
    assert new_records == ["C"]


def test_schema_matching_with_records_in_arrays():
    """https://github.com/fastavro/fastavro/issues/363"""
    original_schema = {
        "type": "record",
        "name": "DataRecord",
        "fields": [
            {
                "name": "string1",
                "type": "string",
            },
            {
                "name": "subrecord",
                "type": {
                    "type": "array",
                    "items": {
                        "type": "record",
                        "name": "SubRecord",
                        "fields": [
                            {
                                "name": "string2",
                                "type": "string",
                            }
                        ],
                    },
                },
            },
        ],
    }

    new_schema = {
        "type": "record",
        "name": "DataRecord",
        "fields": [
            {
                "name": "string1",
                "type": "string",
            },
            {
                "name": "subrecord",
                "type": {
                    "type": "array",
                    "items": {
                        "type": "record",
                        "name": "SubRecord",
                        "fields": [
                            {
                                "name": "string2",
                                "type": "string",
                            },
                            {
                                "name": "logs",
                                "default": None,
                                "type": [
                                    "null",
                                    {
                                        "type": "array",
                                        "items": {
                                            "type": "record",
                                            "name": "LogRecord",
                                            "fields": [
                                                {
                                                    "name": "msg",
                                                    "type": "string",
                                                    "default": "",
                                                }
                                            ],
                                        },
                                    },
                                ],
                            },
                        ],
                    },
                },
            },
        ],
    }

    record = {
        "string1": "test",
        "subrecord": [{"string2": "foo"}],
    }

    binary = avro_to_bytes_with_schema(original_schema, record)

    output_using_original_schema = bytes_with_schema_to_avro(original_schema, binary)
    assert output_using_original_schema == record

    output_using_new_schema = bytes_with_schema_to_avro(new_schema, binary)
    assert output_using_new_schema == {
        "string1": "test",
        "subrecord": [{"string2": "foo", "logs": None}],
    }


def test_schema_migrate_record_to_union():
    """https://github.com/fastavro/fastavro/issues/406"""
    original_schema = {
        "name": "Item",
        "type": "record",
        "fields": [
            {
                "name": "category",
                "type": {
                    "type": "record",
                    "name": "Category",
                    "fields": [{"name": "name", "type": "string"}],
                },
            }
        ],
    }

    new_schema_record_first = {
        "name": "Item",
        "type": "record",
        "fields": [
            {
                "name": "category",
                "type": [
                    {
                        "type": "record",
                        "name": "Category",
                        "fields": [{"name": "name", "type": "string"}],
                    },
                    "null",
                ],
            }
        ],
    }

    new_schema_null_first = {
        "name": "Item",
        "type": "record",
        "fields": [
            {
                "name": "category",
                "type": [
                    "null",
                    {
                        "type": "record",
                        "name": "Category",
                        "fields": [{"name": "name", "type": "string"}],
                    },
                ],
            }
        ],
    }

    record = {"category": {"name": "my-category"}}

    binary = avro_to_bytes_with_schema(original_schema, record)

    output_using_original_schema = bytes_with_schema_to_avro(original_schema, binary)
    assert output_using_original_schema == record

    output_using_new_schema_record_first = bytes_with_schema_to_avro(
        new_schema_record_first, binary
    )
    assert output_using_new_schema_record_first == record

    output_using_new_schema_null_first = bytes_with_schema_to_avro(
        new_schema_null_first, binary
    )
    assert output_using_new_schema_null_first == record


def test_union_of_lists_evolution_with_doc():
    """https://github.com/fastavro/fastavro/issues/486"""
    original_schema = {
        "name": "test_union_of_lists_evolution_with_doc",
        "type": "record",
        "fields": [
            {
                "name": "id",
                "type": [
                    "null",
                    {
                        "name": "some_record",
                        "type": "record",
                        "fields": [{"name": "field", "type": "string"}],
                    },
                ],
            }
        ],
    }

    new_schema = {
        "name": "test_union_of_lists_evolution_with_doc",
        "type": "record",
        "fields": [
            {
                "name": "id",
                "type": [
                    "null",
                    {
                        "name": "some_record",
                        "type": "record",
                        "doc": "some documentation",
                        "fields": [{"name": "field", "type": "string"}],
                    },
                ],
            }
        ],
    }

    record = {"id": {"field": "foo"}}

    binary = avro_to_bytes_with_schema(original_schema, record)

    output_using_new_schema = bytes_with_schema_to_avro(new_schema, binary)
    assert output_using_new_schema == record


def test_union_of_lists_evolution_with_extra_type():
    """https://github.com/fastavro/fastavro/issues/486"""
    original_schema = {
        "name": "test_union_of_lists_evolution_with_extra_type",
        "type": "record",
        "fields": [
            {
                "name": "id",
                "type": [
                    "null",
                    {
                        "name": "some_record",
                        "type": "record",
                        "fields": [{"name": "field", "type": "string"}],
                    },
                ],
            }
        ],
    }

    new_schema = {
        "name": "test_union_of_lists_evolution_with_extra_type",
        "type": "record",
        "fields": [
            {
                "name": "id",
                "type": [
                    "null",
                    {
                        "name": "some_record",
                        "type": "record",
                        "fields": [{"name": "field", "type": "string"}],
                    },
                    "string",
                ],
            }
        ],
    }

    record = {"id": {"field": "foo"}}

    binary = avro_to_bytes_with_schema(original_schema, record)

    output_using_new_schema = bytes_with_schema_to_avro(new_schema, binary)
    assert output_using_new_schema == record



# cython: language_level=3
# cython: auto_cpdef=True

import hashlib
from io import StringIO
from os import path
from copy import deepcopy
import json
from libc.math cimport floor, log10

from ._schema_common import (
    PRIMITIVES,
    UnknownType,
    SchemaParseException,
    RESERVED_PROPERTIES,
    OPTIONAL_FIELD_PROPERTIES,
    RESERVED_FIELD_PROPERTIES,
    JAVA_FINGERPRINT_MAPPING,
    FINGERPRINT_ALGORITHMS,
    RABIN_64,
    rabin_fingerprint,
)


cpdef inline extract_record_type(schema):
    if isinstance(schema, dict):
        return schema["type"]

    if isinstance(schema, list):
        return "union"

    return schema


cpdef inline str extract_logical_type(schema):
    if not isinstance(schema, dict):
        return None
    rt = schema["type"]
    lt = schema.get("logicalType")
    if lt:
        # TODO: Building this string every time is going to be relatively slow.
        return f"{rt}-{lt}"
    return None


cpdef fullname(schema):
    return schema_name(schema, "")[1]


cpdef schema_name(schema, parent_ns):
    try:
        name = schema["name"]
    except KeyError:
        raise SchemaParseException(
            f'"name" is a required field missing from the schema: {schema}'
        )

    namespace = schema.get("namespace", parent_ns)
    if not namespace:
        return namespace, name
    elif "." in name:
        return "", name
    else:
        return namespace, f"{namespace}.{name}"


cpdef expand_schema(schema):
    return parse_schema(schema, expand=True, _write_hint=False)


def parse_schema(
    schema, expand=False, _write_hint=True, _force=False, _named_schemas=None
):
    if _named_schemas is None:
        _named_schemas = {}

    if _force or expand:
        return _parse_schema(
            schema, "", expand, _write_hint, set(), _named_schemas
        )
    elif isinstance(schema, dict) and "__fastavro_parsed" in schema:
        for key, value in schema["__named_schemas"].items():
            _named_schemas[key] = value
        return schema
    elif isinstance(schema, list):
        # If we are given a list we should make sure that the immediate sub
        # schemas have the hint in them
        return [
            parse_schema(s, expand, _write_hint, _force, _named_schemas)
            for s in schema
        ]
    else:
        return _parse_schema(
            schema, "", expand, _write_hint, set(), _named_schemas
        )


cdef _parse_schema(schema, namespace, expand, _write_hint, names, named_schemas):
    # union schemas
    if isinstance(schema, list):
        return [
            _parse_schema(
                s, namespace, expand, False, names, named_schemas
            ) for s in schema
        ]

    # string schemas; this could be either a named schema or a primitive type
    elif not isinstance(schema, dict):
        if schema in PRIMITIVES:
            return schema

        if "." not in schema and namespace:
            schema = namespace + "." + schema

        if schema not in named_schemas:
            raise UnknownType(schema)
        elif expand:
            # If `name` is in the schema, it has been fully resolved and so we
            # can include the full schema. If `name` is not in the schema yet,
            # then we are still recursing that schema and must use the named
            # schema or else we will have infinite recursion when printing the
            # final schema
            if "name" in named_schemas[schema]:
                return named_schemas[schema]
            else:
                return schema
        else:
            return schema

    else:
        # Remaining valid schemas must be dict types
        schema_type = schema["type"]

        parsed_schema = {
            key: value
            for key, value in schema.items()
            if key not in RESERVED_PROPERTIES
        }
        parsed_schema["type"] = schema_type

        if "doc" in schema:
            parsed_schema["doc"] = schema["doc"]

        # Correctness checks for logical types
        logical_type = parsed_schema.get("logicalType")
        if logical_type == "decimal":
            scale = parsed_schema.get("scale")
            if scale and (not isinstance(scale, int) or scale < 0):
                raise SchemaParseException(
                    f"decimal scale must be a postive integer, not {scale}"
                )

            precision = parsed_schema.get("precision")
            if precision:
                if not isinstance(precision, int) or precision <= 0:
                    raise SchemaParseException(
                        "decimal precision must be a postive integer, "
                        + f"not {precision}"
                    )
                if schema_type == "fixed":
                    # https://avro.apache.org/docs/current/spec.html#Decimal
                    size = schema["size"]
                    max_precision = int(floor(log10(2) * (8 * size - 1)))
                    if precision > max_precision:
                        raise SchemaParseException(
                            f"decimal precision of {precision} doesn't fit "
                            + f"into array of length {size}"
                        )

            if scale and precision and precision < scale:
                raise SchemaParseException(
                    "decimal scale must be less than or equal to "
                    + f"the precision of {precision}"
                )

        if schema_type == "array":
            parsed_schema["items"] = _parse_schema(
                schema["items"],
                namespace,
                expand,
                False,
                names,
                named_schemas,
            )

        elif schema_type == "map":
            parsed_schema["values"] = _parse_schema(
                schema["values"],
                namespace,
                expand,
                False,
                names,
                named_schemas,
            )

        elif schema_type == "enum":
            _, fullname = schema_name(schema, namespace)
            if fullname in names:
                raise SchemaParseException(f"redefined named type: {fullname}")
            names.add(fullname)

            named_schemas[fullname] = parsed_schema

            parsed_schema["name"] = fullname
            parsed_schema["symbols"] = schema["symbols"]

        elif schema_type == "fixed":
            _, fullname = schema_name(schema, namespace)
            if fullname in names:
                raise SchemaParseException(f"redefined named type: {fullname}")
            names.add(fullname)

            named_schemas[fullname] = parsed_schema

            parsed_schema["name"] = fullname
            parsed_schema["size"] = schema["size"]

        elif schema_type == "record" or schema_type == "error":
            # records
            namespace, fullname = schema_name(schema, namespace)
            if fullname in names:
                raise SchemaParseException(f"redefined named type: {fullname}")
            names.add(fullname)

            named_schemas[fullname] = parsed_schema

            fields = []
            for field in schema.get("fields", []):
                fields.append(
                    parse_field(field, namespace, expand, names, named_schemas)
                )

            parsed_schema["name"] = fullname
            parsed_schema["fields"] = fields

            # Hint that we have parsed the record
            if _write_hint:
                # Make a copy of parsed_schema so that we don't have a cyclical
                # reference. Using deepcopy is pretty slow, and we don't need a
                # true deepcopy so this works good enough
                named_schemas[fullname] = {
                    k: v for k, v in parsed_schema.items()
                }

                parsed_schema["__fastavro_parsed"] = True
                parsed_schema["__named_schemas"] = named_schemas

        elif schema_type in PRIMITIVES:
            parsed_schema["type"] = schema_type

        else:
            raise UnknownType(schema)

        return parsed_schema


cdef parse_field(field, namespace, expand, names, named_schemas):
    parsed_field = {
        key: value
        for key, value in field.items()
        if key not in RESERVED_FIELD_PROPERTIES
    }

    for prop in OPTIONAL_FIELD_PROPERTIES:
        if prop in field:
            parsed_field[prop] = field[prop]

    # Aliases must be a list
    aliases = parsed_field.get("aliases", [])
    if not isinstance(aliases, list):
        msg = f"aliases must be a list, not {aliases}"
        raise SchemaParseException(msg)

    parsed_field["name"] = field["name"]
    parsed_field["type"] = _parse_schema(
        field["type"], namespace, expand, False, names, named_schemas
    )

    return parsed_field


def load_schema(
    schema_path, *, _named_schemas=None, _write_hint=True, _injected_schemas=None
):
    if _named_schemas is None:
        _named_schemas = {}

    if _injected_schemas is None:
        _injected_schemas = set()

    with open(schema_path) as fd:
        schema = json.load(fd)
    schema_dir, schema_file = path.split(schema_path)
    return _load_schema(
        schema, schema_dir, _named_schemas, _write_hint, _injected_schemas
    )


cdef _load_schema(schema, schema_dir, named_schemas, write_hint, injected_schemas):
    try:
        schema_copy = deepcopy(named_schemas)
        return parse_schema(
            schema, _named_schemas=named_schemas, _write_hint=write_hint
        )
    except UnknownType as e:
        try:
            avsc = path.join(schema_dir, f"{e.name}.avsc")
            sub_schema = load_schema(
                avsc,
                _named_schemas=schema_copy,
                _write_hint=False,
                _injected_schemas=injected_schemas,
            )
        except IOError:
            raise e

        if sub_schema["name"] not in injected_schemas:
            _inject_schema(schema, sub_schema)
            injected_schemas.add(sub_schema["name"])
        return _load_schema(
            schema, schema_dir, schema_copy, write_hint, injected_schemas
        )


cdef _inject_schema(outer_schema, inner_schema, namespace="", is_injected=False):
    # Once injected, we can stop checking to see if we need to inject since it
    # should only be done once at most
    if is_injected is True:
        return outer_schema, is_injected

    # union schemas
    if isinstance(outer_schema, list):
        union = []
        for each_schema in outer_schema:
            if is_injected:
                union.append(each_schema)
            else:
                return_schema, injected = _inject_schema(
                    each_schema, inner_schema, namespace, is_injected
                )
                union.append(return_schema)
                if injected is True:
                    is_injected = injected
        return union, is_injected

    # string schemas; this could be either a named schema or a primitive type
    elif not isinstance(outer_schema, dict):
        if outer_schema in PRIMITIVES:
            return outer_schema, is_injected

        if "." not in outer_schema and namespace:
            outer_schema = namespace + "." + outer_schema

        if outer_schema == inner_schema["name"]:
            return inner_schema, True
        else:
            # Hit a named schema that has already been loaded previously. Return
            # the outer_schema so we keep looking
            return outer_schema, is_injected
    else:
        # Remaining valid schemas must be dict types
        schema_type = outer_schema["type"]

        if schema_type == "array":
            return_schema, injected = _inject_schema(
                outer_schema["items"], inner_schema, namespace, is_injected
            )
            outer_schema["items"] = return_schema
            return outer_schema, injected

        elif schema_type == "map":
            return_schema, injected = _inject_schema(
                outer_schema["values"], inner_schema, namespace, is_injected
            )
            outer_schema["values"] = return_schema
            return outer_schema, injected

        elif schema_type == "enum":
            return outer_schema, is_injected

        elif schema_type == "fixed":
            return outer_schema, is_injected

        elif schema_type == "record" or schema_type == "error":
            # records
            namespace, _ = schema_name(outer_schema, namespace)
            fields = []
            for field in outer_schema.get("fields", []):
                if is_injected:
                    fields.append(field)
                else:
                    return_schema, injected = _inject_schema(
                        field["type"], inner_schema, namespace, is_injected
                    )
                    field["type"] = return_schema
                    fields.append(field)

                    if injected is True:
                        is_injected = injected
            if fields:
                outer_schema["fields"] = fields

            return outer_schema, is_injected

        elif schema_type in PRIMITIVES:
            return outer_schema, is_injected

        else:
            raise Exception(
                "Internal error; "
                + "You should raise an issue in the fastavro github repository"
            )


def load_schema_ordered(ordered_schemas, *, _write_hint=True):
    loaded_schemas = []
    _named_schemas = {}
    for idx, schema_path in enumerate(ordered_schemas):
        # _write_hint is always False except maybe the outer most schema
        _last = _write_hint if idx + 1 == len(ordered_schemas) else False
        schema = load_schema(
            schema_path, _named_schemas=_named_schemas, _write_hint=_last
        )
        loaded_schemas.append(schema)

    top_first_order = loaded_schemas[::-1]
    outer_schema = top_first_order.pop(0)

    while top_first_order:
        sub_schema = top_first_order.pop(0)
        _inject_schema(outer_schema, sub_schema)

    return outer_schema


def to_parsing_canonical_form(schema):
    fo = StringIO()
    _to_parsing_canonical_form(parse_schema(schema), fo)
    return fo.getvalue()


cdef _to_parsing_canonical_form(schema, fo):
    # union schemas
    if isinstance(schema, list):
        fo.write("[")
        for idx, s in enumerate(schema):
            if idx != 0:
                fo.write(",")
            _to_parsing_canonical_form(s, fo)
        fo.write("]")

    # string schemas; this could be either a named schema or a primitive type
    elif not isinstance(schema, dict):
        fo.write(f'"{schema}"')

    else:
        # Remaining valid schemas must be dict types
        schema_type = schema["type"]

        if schema_type == "array":
            fo.write(f'{{"type":"{schema_type}","items":')
            _to_parsing_canonical_form(schema["items"], fo)
            fo.write("}")

        elif schema_type == "map":
            fo.write(f'{{"type":"{schema_type}","values":')
            _to_parsing_canonical_form(schema["values"], fo)
            fo.write("}")

        elif schema_type == "enum":
            name = schema["name"]
            fo.write(f'{{"name":"{name}","type":"{schema_type}","symbols":[')

            for idx, symbol in enumerate(schema["symbols"]):
                if idx != 0:
                    fo.write(",")
                fo.write(f'"{symbol}"')
            fo.write("]}")

        elif schema_type == "fixed":
            name = schema["name"]
            size = schema["size"]
            fo.write(f'{{"name":"{name}","type":"{schema_type}","size":{size}}}')

        elif schema_type == "record" or schema_type == "error":
            name = schema["name"]
            fo.write(f'{{"name":"{name}","type":"record","fields":[')

            for idx, field in enumerate(schema["fields"]):
                if idx != 0:
                    fo.write(",")
                name = field["name"]
                fo.write(f'{{"name":"{name}","type":')
                _to_parsing_canonical_form(field["type"], fo)
                fo.write("}")
            fo.write("]}")

        elif schema_type in PRIMITIVES:
            fo.write(f'"{schema_type}"')


def fingerprint(parsing_canonical_form, algorithm):
    if algorithm not in FINGERPRINT_ALGORITHMS:
        raise ValueError(
            f"Unknown schema fingerprint algorithm {algorithm}. "
            + f"Valid values include: {FINGERPRINT_ALGORITHMS}"
        )

    # Fix Java names
    algorithm = JAVA_FINGERPRINT_MAPPING.get(algorithm, algorithm)

    if algorithm == RABIN_64:
        return rabin_fingerprint(parsing_canonical_form.encode())

    h = hashlib.new(algorithm, parsing_canonical_form.encode())
    return h.hexdigest()



import sys
import pytest
from fastavro._write_common import _is_appendable


def test_appendable_raises_valuerror(tmpdir):
    """_is_appendable() raises ValueError when file is only 'a' mode."""
    test_file = str(tmpdir.join("test.avro"))

    with open(test_file, "a") as new_file:
        new_file.write("this phrase forwards cursor position beyond zero")
        with pytest.raises(ValueError, match=r"you must use the 'a\+' mode"):
            _is_appendable(new_file)


def test_appendable_true_nonzero(tmpdir):
    """_is_appendable() returns True when file_like.tell() is non-zero."""
    test_file = str(tmpdir.join("test.avro"))

    with open(test_file, "a+b") as new_file:
        new_file.write(b"this phrase forwards cursor position beyond zero")
        assert _is_appendable(new_file)


def test_appendable_false_zero():
    """_is_appendable() returns True when file_like.tell() returns 0."""

    class MockFileLike:
        def seekable(self):
            return True

        def tell(self):
            # mock a 0 position
            return 0

    assert not _is_appendable(MockFileLike())


def test_appendable_false_unseekable_stream():
    """File streams that cannot seek return False."""

    class MockStreamLike:
        # This mock file-like object simply returns False for 'seekable', and,
        # if 'tell' were called, rasies OSError. This mimicks a streaming
        # buffer like sys.stdout.buffer without actually using it.
        def seekable(self):
            return False

        def tell(self):
            # mock what a write-only stream would do
            raise OSError(29, "Illegal seek")

    assert not _is_appendable(MockStreamLike())


def test_appendable_false_stdout(capfd):
    """_is_appendable() returns False when file_like is sys.stdout.buffer."""
    # normally, pytest performs "Capturing of stderr/stdout", which is pretty
    # great, but it impacts this "integration test": we'd like to use our true
    # stdout, whether it is a terminal or pipe, to invoke the true behavior of
    # _is_appendable() when used with 'sys.stdout.buffer'.
    with capfd.disabled():
        assert not _is_appendable(sys.stdout.buffer)



# cython: language_level=3

import array
import numbers
from collections.abc import Mapping, Sequence

from . import const
from ._schema import (
    extract_record_type, extract_logical_type, schema_name, parse_schema
)
from ._logical_writers import LOGICAL_WRITERS
from ._schema_common import UnknownType
from ._validate_common import ValidationError, ValidationErrorData

ctypedef int int32
ctypedef unsigned int uint32
ctypedef unsigned long long ulong64
ctypedef long long long64

cdef int32 INT_MIN_VALUE = const.INT_MIN_VALUE
cdef int32 INT_MAX_VALUE = const.INT_MAX_VALUE
cdef long64 LONG_MIN_VALUE = const.LONG_MIN_VALUE
cdef long64 LONG_MAX_VALUE = const.LONG_MAX_VALUE


cdef inline bint validate_null(datum):
    return datum is None


cdef inline bint validate_boolean(datum):
    return isinstance(datum, bool)


cdef inline bint validate_string(datum):
    return isinstance(datum, str)


cdef inline bint validate_bytes(datum):
    return isinstance(datum, (bytes, bytearray))


cdef inline bint validate_int(datum):
    return (
        isinstance(datum, (int, numbers.Integral))
        and INT_MIN_VALUE <= datum <= INT_MAX_VALUE
        and not isinstance(datum, bool)
    )


cdef inline bint validate_long(datum):
    return (
        isinstance(datum, (int, numbers.Integral))
        and LONG_MIN_VALUE <= datum <= LONG_MAX_VALUE
        and not isinstance(datum, bool)
    )


cdef inline bint validate_float(datum):
    return (
        isinstance(datum, (int, float, numbers.Real))
        and not isinstance(datum, bool)
    )


cdef inline bint validate_fixed(datum, dict schema):
    return (
        (isinstance(datum, bytes) or isinstance(datum, bytearray))
        and len(datum) == schema["size"]
    )


cdef inline bint validate_enum(datum, dict schema):
    return datum in schema["symbols"]


cdef inline bint validate_array(
    datum,
    dict schema,
    dict named_schemas,
    str parent_ns="",
    bint raise_errors=True,
) except -1:
    if not isinstance(datum, (Sequence, array.array)) or isinstance(datum, str):
        return False

    for d in datum:
        if not _validate(datum=d, schema=schema["items"],
                         named_schemas=named_schemas,
                         field=parent_ns,
                         raise_errors=raise_errors):
            return False
    return True


cdef inline bint validate_map(
    object datum,
    dict schema,
    dict named_schemas,
    str parent_ns="",
    bint raise_errors=True,
) except -1:
    # initial checks for map type
    if not isinstance(datum, Mapping):
        return False
    for k in datum:
        if not isinstance(k, str):
            return False

    for v in datum.values():
        if not _validate(datum=v, schema=schema["values"],
                         named_schemas=named_schemas,
                         field=parent_ns,
                         raise_errors=raise_errors):
            return False
    return True


cdef inline bint validate_record(
    object datum,
    dict schema,
    dict named_schemas,
    str parent_ns="",
    bint raise_errors=True,
) except -1:
    if not isinstance(datum, Mapping):
        return False
    _, namespace = schema_name(schema, parent_ns)
    for f in schema["fields"]:
        if not _validate(datum=datum.get(f["name"], f.get("default")),
                         schema=f["type"],
                         named_schemas=named_schemas,
                         field=f"{namespace}.{f['name']}",
                         raise_errors=raise_errors):
            return False
    return True


cdef inline bint validate_union(
    object datum,
    list schema,
    dict named_schemas,
    str parent_ns=None,
    bint raise_errors=True,
) except -1:
    if isinstance(datum, tuple):
        (name, datum) = datum
        for candidate in schema:
            if extract_record_type(candidate) == "record":
                schema_name = candidate["name"]
            else:
                schema_name = candidate
            if schema_name == name:
                return _validate(datum, schema=candidate,
                                 named_schemas=named_schemas,
                                 field=parent_ns,
                                 raise_errors=raise_errors)
        else:
            return False

    cdef list errors = []
    for s in schema:
        try:
            ret = _validate(datum, schema=s,
                            named_schemas=named_schemas,
                            field=parent_ns,
                            raise_errors=raise_errors)
            if ret:
                # We exit on the first passing type in Unions
                return True
        except ValidationError as e:
            errors.extend(e.errors)
    if raise_errors:
        raise ValidationError(*errors)
    return False


cpdef _validate(
    object datum,
    object schema,
    dict named_schemas,
    str field="",
    bint raise_errors=True,
):
    record_type = extract_record_type(schema)
    result = None

    logical_type = extract_logical_type(schema)
    if logical_type:
        prepare = LOGICAL_WRITERS.get(logical_type)
        if prepare:
            datum = prepare(datum, schema)

    # explicit, so that cython is faster, but only for Base Validators
    if record_type == "null":
        result = validate_null(datum)
    elif record_type == "boolean":
        result = validate_boolean(datum)
    elif record_type == "string":
        result = validate_string(datum)
    elif record_type == "int":
        result = validate_int(datum)
    elif record_type == "long":
        result = validate_long(datum)
    elif record_type in ("float", "double"):
        result = validate_float(datum)
    elif record_type == "bytes":
        result = validate_bytes(datum)
    elif record_type == "fixed":
        result = validate_fixed(datum, schema=schema)
    elif record_type == "enum":
        result = validate_enum(datum, schema=schema)
    elif record_type == "array":
        result = validate_array(
            datum,
            schema=schema,
            named_schemas=named_schemas,
            parent_ns=field,
            raise_errors=raise_errors,
        )
    elif record_type == "map":
        result = validate_map(
            datum,
            schema=schema,
            named_schemas=named_schemas,
            parent_ns=field,
            raise_errors=raise_errors,
        )
    elif record_type in ("union", "error_union"):
        result = validate_union(
            datum,
            schema=schema,
            named_schemas=named_schemas,
            parent_ns=field,
            raise_errors=raise_errors,
        )
    elif record_type in ("record", "error", "request"):
        result = validate_record(
            datum,
            schema=schema,
            named_schemas=named_schemas,
            parent_ns=field,
            raise_errors=raise_errors,
        )
    elif record_type in named_schemas:
        result = _validate(datum,
                           schema=named_schemas[record_type],
                           named_schemas=named_schemas,
                           field=field,
                           raise_errors=raise_errors)
    else:
        raise UnknownType(record_type)

    if raise_errors and result is False:
        raise ValidationError(ValidationErrorData(datum, schema, field))

    return bool(result)


cpdef validate(object datum, object schema, str field="",
               bint raise_errors=True):
    named_schemas = {}
    parsed_schema = parse_schema(
        schema, _force=True, _named_schemas=named_schemas
    )
    return _validate(datum, parsed_schema, named_schemas, field, raise_errors)


cpdef validate_many(records, schema, bint raise_errors=True):
    cdef bint result
    cdef list errors = []
    cdef list results = []
    named_schemas = {}
    parsed_schema = parse_schema(
        schema, _force=True, _named_schemas=named_schemas
    )
    for record in records:
        try:
            result = _validate(
                record, parsed_schema, named_schemas, raise_errors=raise_errors
            )
            results.append(result)
        except ValidationError as e:
            errors.extend(e.errors)
    if raise_errors and errors:
        raise ValidationError(*errors)
    return all(results)



from io import BytesIO
import fastavro


def roundtrip(record, writer_schema, reader_schema):
    new_file = BytesIO()
    fastavro.writer(new_file, writer_schema, [record])
    new_file.seek(0)

    new_records = list(fastavro.reader(new_file, reader_schema))
    return new_records[0]


def test_int_promotion():
    int_schema = {
        "type": "int",
    }

    long_schema = {
        "type": "long",
    }

    result = roundtrip(1, int_schema, long_schema)
    assert result == 1
    assert isinstance(result, int)

    float_schema = {
        "type": "float",
    }

    result = roundtrip(1, int_schema, float_schema)
    assert result == 1.0
    assert isinstance(result, float)

    double_schema = {
        "type": "double",
    }

    result = roundtrip(1, int_schema, double_schema)
    assert result == 1.0
    # Python doesn't have a double type, but float is close enough
    assert isinstance(result, float)


def test_long_promotion():
    long_schema = {
        "type": "long",
    }

    float_schema = {
        "type": "float",
    }

    result = roundtrip(1, long_schema, float_schema)
    assert result == 1.0
    assert isinstance(result, float)

    double_schema = {
        "type": "double",
    }

    result = roundtrip(1, long_schema, double_schema)
    assert result == 1.0
    # Python doesn't have a double type, but float is close enough
    assert isinstance(result, float)


def test_float_promotion():
    float_schema = {
        "type": "float",
    }

    double_schema = {
        "type": "double",
    }

    result = roundtrip(1.0, float_schema, double_schema)
    assert result == 1.0
    # Python doesn't have a double type, but float is close enough
    assert isinstance(result, float)


def test_string_promotion():
    string_schema = {
        "type": "string",
    }

    bytes_schema = {
        "type": "bytes",
    }

    result = roundtrip("foo", string_schema, bytes_schema)
    assert result == b"foo"
    assert isinstance(result, bytes)


def test_bytes_promotion():
    bytes_schema = {
        "type": "bytes",
    }

    string_schema = {
        "type": "string",
    }

    result = roundtrip(b"foo", bytes_schema, string_schema)
    assert result == "foo"



import decimal
from typing import Any, Callable, Dict, Iterator, IO, Optional, Union, Tuple
from .types import AvroMessage

class reader:
    fo: IO
    return_record_name: bool
    metadata: Dict[str, bytes]
    codec: str
    reader_schema: Optional[Dict]
    writer_schema: Optional[Dict]
    def __init__(
        self, fo: IO, reader_schema: Optional[Dict], return_record_name: bool
    ): ...
    def __iter__(self) -> Iterator[AvroMessage]: ...
    def next(self) -> AvroMessage: ...
    def __next__(self) -> AvroMessage: ...

class block_reader:
    fo: IO
    return_record_name: bool
    metadata: Dict[str, bytes]
    codec: str
    reader_schema: Optional[Dict]
    writer_schema: Optional[Dict]
    def __init__(
        self, fo: IO, reader_schema: Optional[Dict], return_record_name: bool
    ): ...
    def __iter__(self) -> Iterator[Block]: ...
    def next(self) -> Block: ...
    def __next__(self) -> Block: ...

class Block:
    num_records: int
    writer_schema: Dict
    reader_schema: Dict
    offset: int
    size: int
    def __init__(
        self,
        bytes_: bytes,
        num_records: int,
        codec: str,
        reader_schema: Dict,
        writer_schema: Dict,
        offset: int,
        size: int,
        return_record_name: bool,
    ): ...
    def __iter__(self) -> Iterator[AvroMessage]: ...
    def __str__(self) -> str: ...

def json_reader(fo: IO, schema: Dict) -> reader: ...
def schemaless_reader(
    fo: IO, writer_schema: Dict, reader_schema: Optional[Dict], return_record_name: bool
) -> AvroMessage: ...
def is_avro(path_or_buffer: Union[str, IO]) -> bool: ...

logical_reader = Callable[[Any, Optional[Dict], Optional[Dict]], Any]
LOGICAL_READERS: Dict[str, logical_reader]

class SchemaResolutionError(Exception): ...

BLOCK_READERS: Dict[str, Callable]



# cython: language_level=3

import datetime

import decimal
import os
import uuid

from libc.time cimport tm, mktime
from cpython.int cimport PyInt_AS_LONG
from cpython.tuple cimport PyTuple_GET_ITEM

from fastavro import const

ctypedef long long long64

cdef long64 MCS_PER_SECOND = const.MCS_PER_SECOND
cdef long64 MCS_PER_MINUTE = const.MCS_PER_MINUTE
cdef long64 MCS_PER_HOUR = const.MCS_PER_HOUR

cdef long64 MLS_PER_SECOND = const.MLS_PER_SECOND
cdef long64 MLS_PER_MINUTE = const.MLS_PER_MINUTE
cdef long64 MLS_PER_HOUR = const.MLS_PER_HOUR

cdef is_windows = os.name == "nt"
epoch = datetime.datetime(1970, 1, 1, tzinfo=datetime.timezone.utc)
epoch_naive = datetime.datetime(1970, 1, 1)


cpdef prepare_timestamp_millis(object data, schema):
    cdef object tt
    cdef tm time_tuple
    if isinstance(data, datetime.datetime):
        if data.tzinfo is not None:
            delta = data - epoch
            return <long64>(
                (delta.days * 24 * 3600 + delta.seconds) * MLS_PER_SECOND
                + int(delta.microseconds / 1000)
            )
        elif is_windows:
            # On Windows, timestamps before the epoch will raise an error.
            # See https://bugs.python.org/issue36439
            delta = data - epoch_naive
            return <long64>(
                (delta.days * 24 * 3600 + delta.seconds) * MLS_PER_SECOND
                + int(delta.microseconds / 1000)
            )
        else:
            tt = data.timetuple()
            time_tuple.tm_sec = PyInt_AS_LONG(<object>(PyTuple_GET_ITEM(tt, 5)))
            time_tuple.tm_min = PyInt_AS_LONG(<object>(PyTuple_GET_ITEM(tt, 4)))
            time_tuple.tm_hour = PyInt_AS_LONG(<object>(PyTuple_GET_ITEM(tt, 3)))
            time_tuple.tm_mday = PyInt_AS_LONG(<object>(PyTuple_GET_ITEM(tt, 2)))
            time_tuple.tm_mon = PyInt_AS_LONG(<object>(PyTuple_GET_ITEM(tt, 1))) - 1
            time_tuple.tm_year = PyInt_AS_LONG(<object>(PyTuple_GET_ITEM(tt, 0))) - 1900
            time_tuple.tm_isdst = PyInt_AS_LONG(<object>(PyTuple_GET_ITEM(tt, 8)))

            return (
                mktime(& time_tuple) * MLS_PER_SECOND
                + <long64>(int(data.microsecond) / 1000)
            )
    else:
        return data


cpdef prepare_timestamp_micros(object data, schema):
    cdef object tt
    cdef tm time_tuple
    if isinstance(data, datetime.datetime):
        if data.tzinfo is not None:
            delta = data - epoch
            return <long64>(
                (delta.days * 24 * 3600 + delta.seconds) * MCS_PER_SECOND
                + delta.microseconds
            )
        elif is_windows:
            # On Windows, timestamps before the epoch will raise an error.
            # See https://bugs.python.org/issue36439
            delta = data - epoch_naive
            return <long64>(
                (delta.days * 24 * 3600 + delta.seconds) * MCS_PER_SECOND
                + delta.microseconds
            )
        else:
            tt = data.timetuple()
            time_tuple.tm_sec = PyInt_AS_LONG(<object>(PyTuple_GET_ITEM(tt, 5)))
            time_tuple.tm_min = PyInt_AS_LONG(<object>(PyTuple_GET_ITEM(tt, 4)))
            time_tuple.tm_hour = PyInt_AS_LONG(<object>(PyTuple_GET_ITEM(tt, 3)))
            time_tuple.tm_mday = PyInt_AS_LONG(<object>(PyTuple_GET_ITEM(tt, 2)))
            time_tuple.tm_mon = PyInt_AS_LONG(<object>(PyTuple_GET_ITEM(tt, 1))) - 1
            time_tuple.tm_year = PyInt_AS_LONG(<object>(PyTuple_GET_ITEM(tt, 0))) - 1900
            time_tuple.tm_isdst = PyInt_AS_LONG(<object>(PyTuple_GET_ITEM(tt, 8)))

            return mktime(& time_tuple) * MCS_PER_SECOND + <long64>(data.microsecond)
    else:
        return data


cpdef prepare_date(object data, schema):
    if isinstance(data, datetime.date):
        return data.toordinal() - const.DAYS_SHIFT
    elif isinstance(data, str):
        return datetime.datetime.strptime(data, "%Y-%m-%d").toordinal() - const.DAYS_SHIFT
    else:
        return data


cpdef prepare_bytes_decimal(object data, schema):
    """Convert decimal.Decimal to bytes"""
    if not isinstance(data, decimal.Decimal):
        return data
    scale = schema.get("scale", 0)
    precision = schema["precision"]

    sign, digits, exp = data.as_tuple()

    if len(digits) > precision:
        raise ValueError(
            "The decimal precision is bigger than allowed by schema")

    delta = exp + scale

    if delta < 0:
        raise ValueError(
            "Scale provided in schema does not match the decimal")

    unscaled_datum = 0
    for digit in digits:
        unscaled_datum = (unscaled_datum * 10) + digit

    unscaled_datum = 10 ** delta * unscaled_datum

    bytes_req = (unscaled_datum.bit_length() + 8) // 8

    if sign:
        unscaled_datum = -unscaled_datum

    return unscaled_datum.to_bytes(bytes_req, byteorder="big", signed=True)


cpdef prepare_fixed_decimal(object data, schema):
    cdef bytearray tmp
    if not isinstance(data, decimal.Decimal):
        return data
    scale = schema.get("scale", 0)
    size = schema["size"]
    precision = schema["precision"]

    # based on https://github.com/apache/avro/pull/82/

    sign, digits, exp = data.as_tuple()

    if len(digits) > precision:
        raise ValueError(
            "The decimal precision is bigger than allowed by schema")

    if -exp > scale:
        raise ValueError(
            "Scale provided in schema does not match the decimal")

    delta = exp + scale
    if delta > 0:
        digits = digits + (0,) * delta

    unscaled_datum = 0
    for digit in digits:
        unscaled_datum = (unscaled_datum * 10) + digit

    bits_req = unscaled_datum.bit_length() + 1

    size_in_bits = size * 8
    offset_bits = size_in_bits - bits_req

    mask = 2 ** size_in_bits - 1
    bit = 1
    for i in range(bits_req):
        mask ^= bit
        bit <<= 1

    if bits_req < 8:
        bytes_req = 1
    else:
        bytes_req = bits_req // 8
        if bits_req % 8 != 0:
            bytes_req += 1

    tmp = bytearray()

    if sign:
        unscaled_datum = (1 << bits_req) - unscaled_datum
        unscaled_datum = mask | unscaled_datum
        for index in range(size - 1, -1, -1):
            bits_to_write = unscaled_datum >> (8 * index)
            tmp += bytes([bits_to_write & 0xFF])
    else:
        for i in range(offset_bits // 8):
            tmp += bytes([0])
        for index in range(bytes_req - 1, -1, -1):
            bits_to_write = unscaled_datum >> (8 * index)
            tmp += bytes([bits_to_write & 0xFF])

    return tmp


cpdef prepare_uuid(object data, schema):
    if isinstance(data, uuid.UUID):
        return str(data)
    else:
        return data


cpdef prepare_time_millis(object data, schema):
    if isinstance(data, datetime.time):
        return int(
            data.hour * MLS_PER_HOUR + data.minute * MLS_PER_MINUTE
            + data.second * MLS_PER_SECOND + int(data.microsecond / 1000))
    else:
        return data


cpdef prepare_time_micros(object data, schema):
    if isinstance(data, datetime.time):
        return int(data.hour * MCS_PER_HOUR + data.minute * MCS_PER_MINUTE
                   + data.second * MCS_PER_SECOND + data.microsecond)
    else:
        return data


LOGICAL_WRITERS = {
    "long-timestamp-millis": prepare_timestamp_millis,
    "long-timestamp-micros": prepare_timestamp_micros,
    "int-date": prepare_date,
    "bytes-decimal": prepare_bytes_decimal,
    "fixed-decimal": prepare_fixed_decimal,
    "string-uuid": prepare_uuid,
    "int-time-millis": prepare_time_millis,
    "long-time-micros": prepare_time_micros,

}



import decimal
from typing import Union, List, Dict

AvroMessage = Union[
    None,  # 'null' Avro type
    str,  # 'string' and 'enum'
    float,  # 'float' and 'double'
    int,  # 'int' and 'long'
    decimal.Decimal,  # 'fixed'
    bool,  # 'boolean'
    bytes,  # 'bytes'
    List,  # 'array'
    Dict,  # 'map' and 'record'
]



from io import BytesIO
import fastavro


def roundtrip(schema, record):
    new_file = BytesIO()
    fastavro.schemaless_writer(new_file, schema, record)
    new_file.seek(0)
    new_record = fastavro.schemaless_reader(new_file, schema)
    return new_record


def test_schemaless_writer_and_reader():
    schema = {
        "type": "record",
        "name": "Test",
        "namespace": "test",
        "fields": [{"name": "field", "type": {"type": "string"}}],
    }
    record = {"field": "test"}
    assert record == roundtrip(schema, record)


def test_schemaless_writer_and_reader_with_union():
    """Testing basic functionality of reader with union when option to
    return_record_name is true.
    """
    schema = {
        "name": "Message",
        "type": "record",
        "namespace": "test",
        "fields": [
            {"name": "id", "type": "long"},
            {
                "name": "payload",
                "type": [
                    {
                        "name": "ApplicationCreated",
                        "type": "record",
                        "fields": [
                            {"name": "applicationId", "type": "string"},
                            {"name": "data", "type": "string"},
                        ],
                    },
                    {
                        "name": "ApplicationSubmitted",
                        "type": "record",
                        "fields": [
                            {"name": "applicationId", "type": "string"},
                            {"name": "data", "type": "string"},
                        ],
                    },
                ],
            },
        ],
    }
    record = {
        "id": 123,
        "payload": (
            "test.ApplicationSubmitted",
            {"applicationId": "123456789UT", "data": "..."},
        ),
    }
    new_file = BytesIO()
    fastavro.schemaless_writer(new_file, schema, record)
    new_file.seek(0)
    new_record = fastavro.schemaless_reader(new_file, schema, return_record_name=True)
    assert record == new_record


def test_boolean_roundtrip():
    schema = {
        "type": "record",
        "name": "test_boolean_roundtrip",
        "fields": [{"name": "field", "type": "boolean"}],
    }
    record = {"field": True}
    assert record == roundtrip(schema, record)

    record = {"field": False}
    assert record == roundtrip(schema, record)


def test_default_values_in_reader():
    writer_schema = {
        "name": "name1",
        "type": "record",
        "namespace": "namespace1",
        "fields": [{"doc": "test", "type": "int", "name": "good_field"}],
        "doc": "test",
    }

    reader_schema = {
        "name": "name1",
        "doc": "test",
        "namespace": "namespace1",
        "fields": [
            {"name": "good_field", "doc": "test", "type": "int"},
            {
                "name": "good_compatible_field",
                "doc": "test",
                "default": 1,
                "type": "int",
            },
        ],
        "type": "record",
    }

    record = {"good_field": 1}
    new_file = BytesIO()
    fastavro.schemaless_writer(new_file, writer_schema, record)
    new_file.seek(0)
    new_record = fastavro.schemaless_reader(
        new_file,
        writer_schema,
        reader_schema,
    )
    assert new_record == {"good_field": 1, "good_compatible_field": 1}


def test_newer_versions_of_named_schemas():
    """https://github.com/fastavro/fastavro/issues/450"""
    schema_v1 = [
        {
            "name": "Location",
            "type": "record",
            "fields": [{"name": "city", "type": "string"}],
        },
        {
            "name": "Weather",
            "type": "record",
            "fields": [{"name": "of", "type": "Location"}],
        },
    ]

    schema_v2 = [
        {
            "name": "Location",
            "type": "record",
            "fields": [{"name": "city", "type": "long"}],
        },
        {
            "name": "Weather",
            "type": "record",
            "fields": [{"name": "of", "type": "Location"}],
        },
    ]

    example_1 = {"of": {"city": "London"}}
    example_2 = {"of": {"city": 123}}

    parse_v1 = fastavro.parse_schema(schema_v1)
    parse_v2 = fastavro.parse_schema(schema_v2)

    fastavro.schemaless_writer(BytesIO(), parse_v2, example_2)
    fastavro.schemaless_writer(BytesIO(), parse_v1, example_1)


def test_newer_versions_of_named_schemas_2():
    """https://github.com/fastavro/fastavro/issues/450"""
    schema = {
        "name": "Weather",
        "type": "record",
        "fields": [
            {
                "name": "place1",
                "type": {
                    "name": "Location",
                    "type": "record",
                    "fields": [{"name": "city", "type": "string"}],
                },
            },
            {
                "name": "place2",
                "type": "Location",
            },
        ],
    }

    example_1 = {"place1": {"city": "London"}, "place2": {"city": "Berlin"}}
    parsed_schema = fastavro.parse_schema(schema)

    assert example_1 == roundtrip(parsed_schema, example_1)



from copy import deepcopy
from io import StringIO
import json

import pytest

from fastavro import json_writer, json_reader
from fastavro.schema import parse_schema


def roundtrip(schema, records):
    new_file = StringIO()
    json_writer(new_file, schema, records)
    new_file.seek(0)

    new_records = list(json_reader(new_file, schema))
    return new_records


def test_json():
    schema = {
        "type": "record",
        "name": "Test",
        "namespace": "test",
        "fields": [
            {"name": "null", "type": "null"},
            {"name": "boolean", "type": "boolean"},
            {"name": "string", "type": "string"},
            {"name": "bytes", "type": "bytes"},
            {"name": "int", "type": "int"},
            {"name": "long", "type": "long"},
            {"name": "float", "type": "float"},
            {"name": "double", "type": "double"},
            {
                "name": "fixed",
                "type": {"type": "fixed", "name": "fixed_field", "size": 5},
            },
            {
                "name": "union",
                "type": [
                    "null",
                    "int",
                    {
                        "type": "record",
                        "name": "union_record",
                        "fields": [{"name": "union_record_field", "type": "string"}],
                    },
                ],
            },
            {
                "name": "enum",
                "type": {
                    "type": "enum",
                    "name": "enum_field",
                    "symbols": ["FOO", "BAR"],
                },
            },
            {"name": "array", "type": {"type": "array", "items": "string"}},
            {"name": "map", "type": {"type": "map", "values": "int"}},
            {
                "name": "record",
                "type": {
                    "type": "record",
                    "name": "subrecord",
                    "fields": [{"name": "sub_int", "type": "int"}],
                },
            },
        ],
    }

    records = [
        {
            "null": None,
            "boolean": True,
            "string": "foo",
            "bytes": b"\xe2\x99\xa5",
            "int": 1,
            "long": 1 << 33,
            "float": 2.2,
            "double": 3.3,
            "fixed": b"\x61\x61\x61\x61\x61",
            "union": None,
            "enum": "BAR",
            "array": ["a", "b"],
            "map": {"c": 1, "d": 2},
            "record": {"sub_int": 123},
        },
        {
            "null": None,
            "boolean": True,
            "string": "foo",
            "bytes": b"\xe2\x99\xa5",
            "int": 1,
            "long": 1 << 33,
            "float": 2.2,
            "double": 3.3,
            "fixed": b"\x61\x61\x61\x61\x61",
            "union": 321,
            "enum": "BAR",
            "array": ["a", "b"],
            "map": {"c": 1, "d": 2},
            "record": {"sub_int": 123},
        },
        {
            "null": None,
            "boolean": True,
            "string": "foo",
            "bytes": b"\xe2\x99\xa5",
            "int": 1,
            "long": 1 << 33,
            "float": 2.2,
            "double": 3.3,
            "fixed": b"\x61\x61\x61\x61\x61",
            "union": {"union_record_field": "union_field"},
            "enum": "BAR",
            "array": ["a", "b"],
            "map": {"c": 1, "d": 2},
            "record": {"sub_int": 123},
        },
    ]

    new_records = roundtrip(schema, records)
    assert records == new_records


def test_more_than_one_record():
    schema = {
        "type": "record",
        "name": "test_more_than_one_record",
        "namespace": "test",
        "fields": [
            {
                "name": "string",
                "type": "string",
            },
            {
                "name": "int",
                "type": "int",
            },
        ],
    }

    records = [
        {
            "string": "foo",
            "int": 1,
        },
        {
            "string": "bar",
            "int": 2,
        },
    ]

    new_records = roundtrip(schema, records)
    assert records == new_records


def test_encoded_union_output():
    schema = {
        "type": "record",
        "name": "Test",
        "namespace": "test",
        "fields": [
            {
                "name": "union",
                "type": [
                    "null",
                    "int",
                    {
                        "type": "record",
                        "name": "union_record",
                        "fields": [
                            {
                                "name": "union_record_field",
                                "type": "string",
                            }
                        ],
                    },
                ],
            }
        ],
    }

    # A null value is encoded as just null
    records = [{"union": None}]
    new_file = StringIO()
    json_writer(new_file, schema, records)
    assert new_file.getvalue().strip() == json.dumps({"union": None})

    # A non-null, non-named type is encoded as an object with a key for the
    # type
    records = [{"union": 321}]
    new_file = StringIO()
    json_writer(new_file, schema, records)
    assert new_file.getvalue().strip() == json.dumps({"union": {"int": 321}})

    # A non-null, named type is encoded as an object with a key for the name
    records = [{"union": {"union_record_field": "union_field"}}]
    new_file = StringIO()
    json_writer(new_file, schema, records)
    expected = json.dumps(
        {"union": {"test.union_record": {"union_record_field": "union_field"}}}
    )
    assert new_file.getvalue().strip() == expected


def test_union_string_and_bytes():
    schema = {
        "type": "record",
        "name": "Test",
        "namespace": "test",
        "fields": [
            {
                "name": "union",
                "type": [
                    "string",
                    "bytes",
                ],
            }
        ],
    }

    records = [
        {
            "union": "asdf",
        },
        {"union": b"asdf"},
    ]

    new_records = roundtrip(schema, records)
    assert records == new_records


def test_simple_type():
    schema = {"type": "string"}

    records = ["foo", "bar"]

    new_records = roundtrip(schema, records)
    assert records == new_records


def test_array_type_simple():
    schema = {"type": "array", "items": "string"}

    records = [
        ["foo", "bar"],
        ["a", "b"],
    ]

    new_records = roundtrip(schema, records)
    assert records == new_records


def test_array_type_records():
    schema = {
        "type": "array",
        "items": {
            "type": "record",
            "name": "test_array_type",
            "fields": [
                {
                    "name": "field1",
                    "type": "string",
                },
                {
                    "name": "field2",
                    "type": "int",
                },
            ],
        },
    }

    records = [
        [{"field1": "foo", "field2": 1}],
        [{"field1": "bar", "field2": 2}],
    ]

    new_records = roundtrip(schema, records)
    assert records == new_records


def test_empty_maps():
    """https://github.com/fastavro/fastavro/issues/380"""
    schema = {
        "type": "map",
        "values": "int",
    }

    records = [
        {"foo": 1},
        {},
    ]

    new_records = roundtrip(schema, records)
    assert records == new_records


def test_empty_arrays():
    """https://github.com/fastavro/fastavro/issues/380"""
    schema = {
        "type": "array",
        "items": "int",
    }

    records = [
        [1],
        [],
    ]

    new_records = roundtrip(schema, records)
    assert records == new_records


def test_union_in_array():
    """https://github.com/fastavro/fastavro/issues/399"""
    schema = {
        "type": "array",
        "items": [
            {
                "type": "record",
                "name": "rec1",
                "fields": [
                    {
                        "name": "field1",
                        "type": ["string", "null"],
                    }
                ],
            },
            {
                "type": "record",
                "name": "rec2",
                "fields": [
                    {
                        "name": "field2",
                        "type": ["string", "null"],
                    }
                ],
            },
            "null",
        ],
    }

    records = [
        [{"field1": "foo"}, {"field2": None}, None],
    ]

    new_records = roundtrip(schema, records)
    assert records == new_records


def test_union_in_array2():
    """https://github.com/fastavro/fastavro/issues/399"""
    schema = {
        "type": "record",
        "name": "Inbox",
        "fields": [
            {"type": "string", "name": "id"},
            {"type": "string", "name": "msg_title"},
            {
                "name": "msg_content",
                "type": {
                    "type": "array",
                    "items": [
                        {
                            "type": "record",
                            "name": "LimitedTime",
                            "fields": [
                                {
                                    "type": ["string", "null"],
                                    "name": "type",
                                    "default": "now",
                                }
                            ],
                        },
                        {
                            "type": "record",
                            "name": "Text",
                            "fields": [{"type": ["string", "null"], "name": "text"}],
                        },
                    ],
                },
            },
        ],
    }

    records = [
        {
            "id": 1234,
            "msg_title": "Hi",
            "msg_content": [{"type": "now"}, {"text": "hi from here!"}],
        },
    ]

    new_records = roundtrip(schema, records)
    assert records == new_records


def test_union_in_map():
    """https://github.com/fastavro/fastavro/issues/399"""
    schema = {
        "type": "record",
        "name": "Test",
        "namespace": "test",
        "fields": [
            {
                "name": "map",
                "type": {
                    "type": "map",
                    "values": ["string", "null"],
                },
            }
        ],
    }

    records = [{"map": {"c": "1", "d": None}}]

    new_records = roundtrip(schema, records)
    assert records == new_records


def test_with_dependent_schema():
    """Tests a schema with dependent schema
    https://github.com/fastavro/fastavro/issues/418"""
    dependency = {
        "type": "record",
        "name": "Dependency",
        "namespace": "test",
        "fields": [{"name": "_name", "type": "string"}],
    }

    schema = {
        "type": "record",
        "name": "Test",
        "namespace": "test",
        "fields": [
            {"name": "_name", "type": "string"},
            {"name": "_dependency", "type": "Dependency"},
        ],
    }

    records = [{"_name": "parent", "_dependency": {"_name": "child"}}]

    parsed_schema = parse_schema([dependency, schema])

    new_records = roundtrip(parsed_schema, records)
    assert records == new_records


def test_enum_named_type():
    """https://github.com/fastavro/fastavro/issues/450"""
    schema = {
        "type": "record",
        "name": "test_enum_named_type",
        "fields": [
            {
                "name": "test1",
                "type": {
                    "type": "enum",
                    "name": "my_enum",
                    "symbols": ["FOO", "BAR"],
                },
            },
            {
                "name": "test2",
                "type": "my_enum",
            },
        ],
    }

    records = [{"test1": "FOO", "test2": "BAR"}]
    parsed_schema = parse_schema(schema)
    assert records == roundtrip(parsed_schema, records)


def test_fixed_named_type():
    """https://github.com/fastavro/fastavro/issues/450"""
    schema = {
        "type": "record",
        "name": "test_fixed_named_type",
        "fields": [
            {
                "name": "test1",
                "type": {
                    "type": "fixed",
                    "name": "my_fixed",
                    "size": 4,
                },
            },
            {
                "name": "test2",
                "type": "my_fixed",
            },
        ],
    }

    records = [{"test1": b"1234", "test2": b"4321"}]
    parsed_schema = parse_schema(schema)
    assert records == roundtrip(parsed_schema, records)


def test_record_named_type():
    """https://github.com/fastavro/fastavro/issues/450"""
    schema = {
        "type": "record",
        "name": "test_record_named_type",
        "fields": [
            {
                "name": "test1",
                "type": {
                    "type": "record",
                    "name": "my_record",
                    "fields": [
                        {
                            "name": "field1",
                            "type": "string",
                        }
                    ],
                },
            },
            {
                "name": "test2",
                "type": "my_record",
            },
        ],
    }

    records = [{"test1": {"field1": "foo"}, "test2": {"field1": "bar"}}]
    parsed_schema = parse_schema(schema)
    assert records == roundtrip(parsed_schema, records)


def test_default_union_values():
    """https://github.com/fastavro/fastavro/issues/485"""
    schema = {
        "type": "record",
        "name": "User",
        "fields": [
            {"name": "name", "type": "string"},
            {"name": "age", "type": "long"},
            {
                "name": "pets",
                "type": {"type": "array", "items": "string"},
            },
            {
                "name": "accounts",
                "type": {"type": "map", "values": "long"},
            },
            {
                "name": "favorite_colors",
                "type": {
                    "type": "enum",
                    "name": "favorite_color",
                    "symbols": ["BLUE", "YELLOW", "GREEN"],
                },
            },
            {"name": "country", "type": ["string", "null"], "default": "Argentina"},
            {"name": "address", "type": ["null", "string"], "default": None},
        ],
        "doc": "An User",
        "namespace": "User.v1",
        "aliases": ["user-v1", "super user"],
    }

    record = {
        "name": "MgXqfDAqzbgJSTTHDXtN",
        "age": 551,
        "pets": ["aRvwODwbOWfrkxYYkJiI"],
        "accounts": {"DQSZRzofFrNCiOhhIOvX": 4431},
        "favorite_colors": "GREEN",
        "address": {"string": "YgmVDKhXctMgODKkhNHJ"},
    }

    new_file = StringIO(json.dumps(record))
    read_record = next(json_reader(new_file, schema))

    assert read_record["country"] == "Argentina"


def test_all_default_values():
    """https://github.com/fastavro/fastavro/issues/485"""
    default_boolean = True
    default_string = "default_string"
    default_bytes = "default_bytes"
    default_int = -1
    default_long = -2
    default_float = 1.1
    default_double = 2.2
    default_fixed = "12345"
    default_union = None
    default_enum = "FOO"
    default_array = ["a", "b"]
    default_map = {"a": 1, "b": 2}
    default_record = {"sub_int": -3}
    schema = {
        "type": "record",
        "name": "test_all_default_values",
        "fields": [
            {"name": "boolean", "type": "boolean", "default": default_boolean},
            {"name": "string", "type": "string", "default": default_string},
            {"name": "bytes", "type": "bytes", "default": default_bytes},
            {"name": "int", "type": "int", "default": default_int},
            {"name": "long", "type": "long", "default": default_long},
            {"name": "float", "type": "float", "default": default_float},
            {"name": "double", "type": "double", "default": default_double},
            {
                "name": "fixed",
                "type": {"type": "fixed", "name": "fixed_field", "size": 5},
                "default": default_fixed,
            },
            {
                "name": "union",
                "type": [
                    "null",
                    "int",
                    {
                        "type": "record",
                        "name": "union_record",
                        "fields": [{"name": "union_record_field", "type": "string"}],
                    },
                ],
                "default": default_union,
            },
            {
                "name": "enum",
                "type": {
                    "type": "enum",
                    "name": "enum_field",
                    "symbols": ["FOO", "BAR"],
                },
                "default": default_enum,
            },
            {
                "name": "array",
                "type": {"type": "array", "items": "string"},
                "default": deepcopy(default_array),
            },
            {
                "name": "map",
                "type": {"type": "map", "values": "int"},
                "default": deepcopy(default_map),
            },
            {
                "name": "record",
                "type": {
                    "type": "record",
                    "name": "subrecord",
                    "fields": [{"name": "sub_int", "type": "int"}],
                },
                "default": default_record,
            },
        ],
    }

    record = {}

    new_file = StringIO(json.dumps(record))
    read_record = next(json_reader(new_file, schema))

    assert read_record["boolean"] == default_boolean
    assert read_record["string"] == default_string
    assert read_record["bytes"] == default_bytes.encode("iso-8859-1")
    assert read_record["int"] == default_int
    assert read_record["long"] == default_long
    assert read_record["float"] == default_float
    assert read_record["double"] == default_double
    assert read_record["fixed"] == default_fixed.encode("iso-8859-1")
    assert read_record["union"] == default_union
    assert read_record["enum"] == default_enum
    assert read_record["array"] == default_array
    assert read_record["map"] == default_map
    assert read_record["record"] == default_record


def test_default_value_missing():
    """https://github.com/fastavro/fastavro/issues/485"""
    schema = {
        "type": "record",
        "name": "test_default_value_missing",
        "fields": [{"name": "string", "type": "string"}],
    }

    record = {}

    new_file = StringIO(json.dumps(record))
    with pytest.raises(ValueError, match="no value and no default"):
        next(json_reader(new_file, schema))



from io import BytesIO
from os import SEEK_SET
from random import choice, seed
from string import ascii_uppercase, digits

import fastavro


def test_str_py3():
    letters = ascii_uppercase + digits
    id_size = 100

    seed("str_py3")  # Repeatable results

    def gen_id():
        return "".join(choice(letters) for _ in range(id_size))

    keys = ["first", "second", "third", "fourth"]

    testdata = [{key: gen_id() for key in keys} for _ in range(50)]

    schema = {
        "fields": [{"name": key, "type": "string"} for key in keys],
        "namespace": "namespace",
        "name": "zerobyte",
        "type": "record",
    }

    buf = BytesIO()
    fastavro.writer(buf, schema, testdata)

    buf.seek(0, SEEK_SET)
    for i, rec in enumerate(fastavro.reader(buf), 1):
        pass

    size = len(testdata)

    assert i == size, "bad number of records"
    assert rec == testdata[-1], "bad last record"


def test_py3_union_string_and_bytes():
    schema = {
        "fields": [{"name": "field", "type": ["string", "bytes"]}],
        "namespace": "namespace",
        "name": "union_string_bytes",
        "type": "record",
    }

    records = [{"field": "string"}, {"field": b"bytes"}]

    buf = BytesIO()
    fastavro.writer(buf, schema, records)



import ast
import os
import re
import sys

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

from setuptools import Extension

# publish.sh should set this variable to 1.
try:
    USE_CYTHON = int(os.getenv("FASTAVRO_USE_CYTHON"))
except TypeError:
    USE_CYTHON = False

ext = ".pyx" if USE_CYTHON else ".c"

# See http://setuptools.readthedocs.io/en/latest/setuptools.html\
#     #distributing-extensions-compiled-with-pyrex
ext_modules = []
if not hasattr(sys, "pypy_version_info"):
    ext_modules += [
        Extension("fastavro._read", ["fastavro/_read" + ext]),
        Extension("fastavro._schema", ["fastavro/_schema" + ext]),
        Extension("fastavro._write", ["fastavro/_write" + ext]),
        Extension("fastavro._validation", ["fastavro/_validation" + ext]),
        Extension("fastavro._logical_writers", ["fastavro/_logical_writers" + ext]),
    ]


def version():
    pyfile = "fastavro/__init__.py"
    with open(pyfile) as fp:
        data = fp.read()

    match = re.search("__version_info__ = (\(.*\))", data)
    assert match, f"cannot find version in {pyfile}"
    vinfo = ast.literal_eval(match.group(1))
    return ".".join(str(v) for v in vinfo)


setup_requires = []
if not hasattr(sys, "pypy_version_info"):
    cpython_requires = [
        # Setuptools 18.0 properly handles Cython extensions.
        "setuptools>=18.0",
    ]
    if USE_CYTHON:
        cpython_requires += [
            "Cython",
        ]
    setup_requires += cpython_requires

tests_require = ["pytest", "flake8", "check-manifest"]
if sys.implementation.name != "pypy":
    tests_require.append("mypy")

setup(
    name="fastavro",
    version=version(),
    description="Fast read/write of AVRO files",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Miki Tebeka",
    author_email="miki.tebeka@gmail.com",
    license="MIT",
    url="https://github.com/fastavro/fastavro",
    packages=["fastavro", "fastavro.io"],
    ext_modules=ext_modules,
    zip_safe=False,
    entry_points={
        "console_scripts": [
            "fastavro = fastavro.__main__:main",
        ]
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: Libraries",
        "Topic :: Scientific/Engineering :: Information Analysis",
    ],
    python_requires=">=3.6",
    extras_require={
        "codecs": ["python-snappy", "zstandard", "lz4"],
        "snappy": ["python-snappy"],
        "zstandard": ["zstandard"],
        "lz4": ["lz4"],
    },
    tests_require=tests_require,
    setup_requires=setup_requires,
    package_data={"fastavro": ["py.typed"]},
)



from io import BytesIO
import os

import pytest

import fastavro


@pytest.mark.parametrize("codec", ["null", "deflate", "bzip2", "xz"])
def test_builtin_codecs(codec):
    schema = {
        "doc": "A weather reading.",
        "name": "Weather",
        "namespace": "test",
        "type": "record",
        "fields": [
            {"name": "station", "type": "string"},
            {"name": "time", "type": "long"},
            {"name": "temp", "type": "int"},
        ],
    }

    records = [
        {"station": "011990-99999", "temp": 0, "time": 1433269388},
        {"station": "011990-99999", "temp": 22, "time": 1433270389},
        {"station": "011990-99999", "temp": -11, "time": 1433273379},
        {"station": "012650-99999", "temp": 111, "time": 1433275478},
    ]

    file = BytesIO()
    fastavro.writer(file, schema, records, codec=codec)

    file.seek(0)
    out_records = list(fastavro.reader(file))
    assert records == out_records


@pytest.mark.parametrize("codec", ["snappy", "zstandard", "lz4"])
@pytest.mark.skipif(os.name == "nt", reason="A pain to set up on windows")
def test_optional_codecs(codec):
    schema = {
        "doc": "A weather reading.",
        "name": "Weather",
        "namespace": "test",
        "type": "record",
        "fields": [
            {"name": "station", "type": "string"},
            {"name": "time", "type": "long"},
            {"name": "temp", "type": "int"},
        ],
    }

    records = [
        {"station": "011990-99999", "temp": 0, "time": 1433269388},
        {"station": "011990-99999", "temp": 22, "time": 1433270389},
        {"station": "011990-99999", "temp": -11, "time": 1433273379},
        {"station": "012650-99999", "temp": 111, "time": 1433275478},
    ]

    file = BytesIO()
    fastavro.writer(file, schema, records, codec=codec)

    file.seek(0)
    out_records = list(fastavro.reader(file))
    assert records == out_records


@pytest.mark.parametrize("codec", ["snappy", "zstandard", "lz4"])
@pytest.mark.skipif(os.name != "nt", reason="codec is present")
def test_optional_codecs_not_installed(codec):
    schema = {
        "doc": "A weather reading.",
        "name": "Weather",
        "namespace": "test",
        "type": "record",
        "fields": [
            {"name": "station", "type": "string"},
            {"name": "time", "type": "long"},
            {"name": "temp", "type": "int"},
        ],
    }

    records = [
        {"station": "011990-99999", "temp": 0, "time": 1433269388},
        {"station": "011990-99999", "temp": 22, "time": 1433270389},
        {"station": "011990-99999", "temp": -11, "time": 1433273379},
        {"station": "012650-99999", "temp": 111, "time": 1433275478},
    ]

    file = BytesIO()
    with pytest.raises(
        ValueError, match=f"{codec} codec is supported but you need to install"
    ):
        fastavro.writer(file, schema, records, codec=codec)


def test_unsupported_codec():
    schema = {
        "doc": "A weather reading.",
        "name": "Weather",
        "namespace": "test",
        "type": "record",
        "fields": [
            {"name": "station", "type": "string"},
            {"name": "time", "type": "long"},
            {"name": "temp", "type": "int"},
        ],
    }

    records = [
        {"station": "011990-99999", "temp": 0, "time": 1433269388},
        {"station": "011990-99999", "temp": 22, "time": 1433270389},
        {"station": "011990-99999", "temp": -11, "time": 1433273379},
        {"station": "012650-99999", "temp": 111, "time": 1433275478},
    ]

    file = BytesIO()
    with pytest.raises(ValueError, match="unrecognized codec"):
        fastavro.writer(file, schema, records, codec="unsupported")

    file = BytesIO()
    fastavro.writer(file, schema, records, codec="deflate")

    # Change the avro binary to act as if it were written with a codec called
    # `unsupported`
    modified_avro = file.getvalue().replace(b"\x0edeflate", b"\x16unsupported")
    modified_file = BytesIO(modified_avro)

    with pytest.raises(ValueError, match="Unrecognized codec"):
        list(fastavro.reader(modified_file))


def test_compression_level():
    """https://github.com/fastavro/fastavro/issues/377"""
    schema = {
        "doc": "A weather reading.",
        "name": "Weather",
        "namespace": "test",
        "type": "record",
        "fields": [
            {"name": "station", "type": "string"},
            {"name": "time", "type": "long"},
            {"name": "temp", "type": "int"},
        ],
    }

    records = [
        {"station": "011990-99999", "temp": 0, "time": 1433269388},
        {"station": "011990-99999", "temp": 22, "time": 1433270389},
        {"station": "011990-99999", "temp": -11, "time": 1433273379},
        {"station": "012650-99999", "temp": 111, "time": 1433275478},
    ]

    file = BytesIO()
    fastavro.writer(file, schema, records, codec="deflate", codec_compression_level=9)

    file.seek(0)
    out_records = list(fastavro.reader(file))
    assert records == out_records



from os.path import join, abspath, dirname
import pytest
import fastavro
from fastavro.schema import (
    SchemaParseException,
    UnknownType,
    parse_schema,
    fullname,
    expand_schema,
)


def test_named_types_have_names():
    record_schema = {
        "type": "record",
        "fields": [
            {
                "name": "field",
                "type": "string",
            }
        ],
    }

    with pytest.raises(SchemaParseException):
        fastavro.parse_schema(record_schema)

    error_schema = {
        "type": "error",
        "fields": [
            {
                "name": "field",
                "type": "string",
            }
        ],
    }

    with pytest.raises(SchemaParseException):
        fastavro.parse_schema(error_schema)

    fixed_schema = {
        "type": "fixed",
        "size": 1,
    }

    with pytest.raises(SchemaParseException):
        fastavro.parse_schema(fixed_schema)

    enum_schema = {
        "type": "enum",
        "symbols": ["FOO"],
    }

    with pytest.raises(SchemaParseException):
        fastavro.parse_schema(enum_schema)

    # Should parse with name
    for schema in (record_schema, error_schema, fixed_schema, enum_schema):
        schema["name"] = "test_named_types_have_names"
        fastavro.parse_schema(schema)


def test_parse_schema():
    schema = {
        "type": "record",
        "name": "test_parse_schema",
        "fields": [
            {
                "name": "field",
                "type": "string",
            }
        ],
    }

    parsed_schema = parse_schema(schema)
    assert "__fastavro_parsed" in parsed_schema

    parsed_schema_again = parse_schema(parsed_schema)
    assert parsed_schema_again == parsed_schema


def test_unknown_type():
    schema = {
        "type": "unknown",
    }

    with pytest.raises(UnknownType):
        parse_schema(schema)


def test_aliases_are_preserved():
    schema = {
        "type": "record",
        "name": "test_parse_schema",
        "fields": [
            {
                "name": "field",
                "type": "string",
                "aliases": ["test"],
            }
        ],
    }

    parsed_schema = parse_schema(schema)
    assert "aliases" in parsed_schema["fields"][0]


def test_aliases_is_a_list():
    """https://github.com/fastavro/fastavro/issues/206"""
    schema = {
        "type": "record",
        "name": "test_parse_schema",
        "fields": [
            {
                "name": "field",
                "type": "string",
                "aliases": "foobar",
            }
        ],
    }

    with pytest.raises(SchemaParseException):
        parse_schema(schema)


def test_decimal_scale_is_an_int():
    """https://github.com/fastavro/fastavro/issues/262"""
    schema = {
        "type": "record",
        "name": "test_scale_is_an_int",
        "fields": [
            {
                "name": "field",
                "type": {
                    "logicalType": "decimal",
                    "precision": 5,
                    "scale": "2",
                    "type": "bytes",
                },
            }
        ],
    }

    with pytest.raises(
        SchemaParseException, match="decimal scale must be a postive integer"
    ):
        parse_schema(schema)


def test_decimal_scale_is_a_positive_int():
    """https://github.com/fastavro/fastavro/issues/457"""
    schema = {
        "type": "record",
        "name": "test_scale_is_an_int",
        "fields": [
            {
                "name": "field",
                "type": {
                    "logicalType": "decimal",
                    "precision": 5,
                    "scale": -2,
                    "type": "bytes",
                },
            }
        ],
    }

    with pytest.raises(
        SchemaParseException, match="decimal scale must be a postive integer"
    ):
        parse_schema(schema)


def test_decimal_precision_is_an_int():
    """https://github.com/fastavro/fastavro/issues/262"""
    schema = {
        "type": "record",
        "name": "test_scale_is_an_int",
        "fields": [
            {
                "name": "field",
                "type": {
                    "logicalType": "decimal",
                    "precision": "5",
                    "scale": 2,
                    "type": "bytes",
                },
            }
        ],
    }

    with pytest.raises(
        SchemaParseException,
        match="decimal precision must be a postive integer",
    ):
        parse_schema(schema)


def test_decimal_precision_is_a_positive_int():
    """https://github.com/fastavro/fastavro/issues/457"""
    schema = {
        "type": "record",
        "name": "test_scale_is_an_int",
        "fields": [
            {
                "name": "field",
                "type": {
                    "logicalType": "decimal",
                    "precision": -5,
                    "scale": 2,
                    "type": "bytes",
                },
            }
        ],
    }

    with pytest.raises(
        SchemaParseException,
        match="decimal precision must be a postive integer",
    ):
        parse_schema(schema)


def test_decimal_precision_is_greater_than_scale():
    """https://github.com/fastavro/fastavro/issues/457"""
    schema = {
        "type": "record",
        "name": "test_scale_is_an_int",
        "fields": [
            {
                "name": "field",
                "type": {
                    "logicalType": "decimal",
                    "precision": 5,
                    "scale": 10,
                    "type": "bytes",
                },
            }
        ],
    }

    with pytest.raises(
        SchemaParseException,
        match="decimal scale must be less than or equal to",
    ):
        parse_schema(schema)


def test_decimal_fixed_accommodates_precision():
    """https://github.com/fastavro/fastavro/issues/457"""
    schema = {
        "type": "record",
        "name": "test_scale_is_an_int",
        "fields": [
            {
                "name": "field",
                "type": {
                    "name": "fixed_decimal",
                    "logicalType": "decimal",
                    "precision": 10,
                    "scale": 2,
                    "type": "fixed",
                    "size": 2,
                },
            }
        ],
    }

    with pytest.raises(
        SchemaParseException,
        match=r"decimal precision of \d+ doesn't fit into array of length \d+",
    ):
        parse_schema(schema)


def test_named_type_cannot_be_redefined():
    schema = {
        "type": "record",
        "namespace": "test.avro.training",
        "name": "SomeMessage",
        "fields": [
            {
                "name": "is_error",
                "type": "boolean",
                "default": False,
            },
            {
                "name": "outcome",
                "type": [
                    {
                        "type": "record",
                        "name": "SomeMessage",
                        "fields": [],
                    },
                    {
                        "type": "record",
                        "name": "ErrorRecord",
                        "fields": [
                            {
                                "name": "errors",
                                "type": {"type": "map", "values": "string"},
                                "doc": "doc",
                            }
                        ],
                    },
                ],
            },
        ],
    }

    with pytest.raises(
        SchemaParseException,
        match="redefined named type: test.avro.training.SomeMessage",
    ):
        parse_schema(schema)

    schema = {
        "type": "record",
        "name": "SomeMessage",
        "fields": [
            {
                "name": "field1",
                "type": {
                    "type": "record",
                    "name": "ThisName",
                    "fields": [],
                },
            },
            {
                "name": "field2",
                "type": {
                    "type": "enum",
                    "name": "ThisName",
                    "symbols": ["FOO", "BAR"],
                },
            },
        ],
    }

    with pytest.raises(SchemaParseException, match="redefined named type: ThisName"):
        parse_schema(schema)

    schema = {
        "type": "record",
        "name": "SomeMessage",
        "fields": [
            {
                "name": "field1",
                "type": {
                    "type": "record",
                    "name": "ThatName",
                    "fields": [],
                },
            },
            {
                "name": "field2",
                "type": {
                    "type": "fixed",
                    "name": "ThatName",
                    "size": 8,
                },
            },
        ],
    }

    with pytest.raises(SchemaParseException, match="redefined named type: ThatName"):
        parse_schema(schema)


def test_doc_left_in_parse_schema():
    schema = {
        "type": "record",
        "name": "test_doc_left_in_parse_schema",
        "doc": "blah",
        "fields": [{"name": "field1", "type": "string", "default": ""}],
    }
    assert schema == parse_schema(schema, _write_hint=False)


def test_schema_fullname_api():
    schema = {
        "type": "record",
        "namespace": "namespace",
        "name": "test_schema_fullname_api",
        "fields": [],
    }

    assert fullname(schema) == "namespace.test_schema_fullname_api"


def test_schema_expansion():
    """https://github.com/fastavro/fastavro/issues/314"""
    sub_schema = {
        "name": "Dependency",
        "namespace": "com.namespace.dependencies",
        "type": "record",
        "fields": [{"name": "sub_field_1", "type": "string"}],
    }

    outer_schema = {
        "name": "MasterSchema",
        "namespace": "com.namespace.master",
        "type": "record",
        "fields": [
            {"name": "field_1", "type": "com.namespace.dependencies.Dependency"}
        ],
    }

    combined = {
        "name": "com.namespace.master.MasterSchema",
        "type": "record",
        "fields": [
            {
                "name": "field_1",
                "type": {
                    "name": "com.namespace.dependencies.Dependency",
                    "type": "record",
                    "fields": [{"name": "sub_field_1", "type": "string"}],
                },
            }
        ],
    }

    parsed = expand_schema([sub_schema, outer_schema])

    assert parsed[1] == combined


def test_schema_expansion_2():
    """https://github.com/fastavro/fastavro/issues/314"""
    original_schema = {
        "name": "MasterSchema",
        "namespace": "com.namespace.master",
        "type": "record",
        "fields": [
            {
                "name": "field_1",
                "type": {
                    "name": "Dependency",
                    "namespace": "com.namespace.dependencies",
                    "type": "record",
                    "fields": [{"name": "sub_field_1", "type": "string"}],
                },
            },
            {"name": "field_2", "type": "com.namespace.dependencies.Dependency"},
        ],
    }

    expanded_schema = {
        "name": "com.namespace.master.MasterSchema",
        "type": "record",
        "fields": [
            {
                "name": "field_1",
                "type": {
                    "name": "com.namespace.dependencies.Dependency",
                    "type": "record",
                    "fields": [{"name": "sub_field_1", "type": "string"}],
                },
            },
            {
                "name": "field_2",
                "type": {
                    "name": "com.namespace.dependencies.Dependency",
                    "type": "record",
                    "fields": [{"name": "sub_field_1", "type": "string"}],
                },
            },
        ],
    }

    assert expanded_schema == expand_schema(original_schema)


def test_expanding_recursive_schemas_should_stop():
    """https://github.com/fastavro/fastavro/issues/314"""
    sub_schema = {
        "name": "LongList",
        "type": "record",
        "fields": [
            {"name": "value", "type": "long"},
            {"name": "next", "type": ["LongList", "null"]},
        ],
    }

    parsed = expand_schema(sub_schema)
    assert sub_schema == parsed


def test_parse_schema_includes_hint_with_list():
    """https://github.com/fastavro/fastavro/issues/444"""
    schema = [
        {
            "type": "record",
            "name": "test_parse_schema_includes_hint_with_list_1",
            "doc": "blah",
            "fields": [{"name": "field1", "type": "string", "default": ""}],
        },
        {
            "type": "record",
            "name": "test_parse_schema_includes_hint_with_list_2",
            "doc": "blah",
            "fields": [{"name": "field2", "type": "string", "default": ""}],
        },
    ]
    parsed_schema = parse_schema(schema)
    for s in parsed_schema:
        assert "__fastavro_parsed" in s


def test_union_schemas_must_have_names_in_order():
    """https://github.com/fastavro/fastavro/issues/450"""
    schema1 = [
        {
            "name": "Location",
            "type": "record",
            "fields": [{"name": "city", "type": "long"}],
        },
        {
            "name": "Weather",
            "type": "record",
            "fields": [{"name": "of", "type": "Location"}],
        },
    ]
    # This should work because Location is defined first
    parse_schema(schema1)

    schema2 = [
        {
            "name": "Weather",
            "type": "record",
            "fields": [{"name": "of", "type": "Location"}],
        },
        {
            "name": "Location",
            "type": "record",
            "fields": [{"name": "city", "type": "long"}],
        },
    ]
    # This should not work because Location is defined after it is used
    with pytest.raises(UnknownType):
        parse_schema(schema2)


def test_load_schema_does_not_make_unions_of_unions():
    """https://github.com/fastavro/fastavro/issues/443"""
    load_schema_dir = join(abspath(dirname(__file__)), "load_schema_test")
    schema_path = join(load_schema_dir, "A.avsc")
    loaded_schema = fastavro.schema.load_schema(schema_path)
    assert isinstance(loaded_schema, dict)


def test_load_schema_does_not_make_unions_of_unions_2():
    """https://github.com/fastavro/fastavro/issues/443"""
    load_schema_dir = join(abspath(dirname(__file__)), "load_schema_test_2")
    schema_path = join(load_schema_dir, "A.avsc")
    loaded_schema = fastavro.schema.load_schema(schema_path)
    assert isinstance(loaded_schema, list)
    for schema in loaded_schema:
        assert not isinstance(schema, list)


def test_load_schema_output_is_correct():
    """https://github.com/fastavro/fastavro/issues/476"""
    load_schema_dir = join(abspath(dirname(__file__)), "load_schema_test")
    schema_path = join(load_schema_dir, "A.avsc")
    loaded_schema = fastavro.schema.load_schema(schema_path, _write_hint=False)

    expected_schema = {
        "name": "A",
        "type": "record",
        "fields": [
            {
                "name": "b",
                "type": [
                    "null",
                    {
                        "name": "B",
                        "type": "record",
                        "fields": [
                            {
                                "name": "c",
                                "type": [
                                    "null",
                                    {
                                        "name": "C",
                                        "type": "record",
                                        "fields": [{"name": "foo", "type": "string"}],
                                    },
                                ],
                                "default": None,
                            }
                        ],
                    },
                ],
                "default": None,
            },
            {
                "name": "d",
                "type": [
                    "null",
                    {
                        "name": "D",
                        "type": "record",
                        "fields": [{"name": "bar", "type": "string"}],
                    },
                ],
                "default": None,
            },
        ],
    }
    assert loaded_schema == expected_schema


def test_load_schema_output_is_correct_3():
    """https://github.com/fastavro/fastavro/issues/476"""
    load_schema_dir = join(abspath(dirname(__file__)), "load_schema_test_3")
    schema_path = join(load_schema_dir, "A.avsc")
    loaded_schema = fastavro.schema.load_schema(schema_path, _write_hint=False)

    expected_schema = {
        "name": "A",
        "type": "record",
        "fields": [
            {
                "name": "b",
                "type": [
                    "null",
                    {
                        "name": "B",
                        "type": "record",
                        "fields": [
                            {
                                "name": "c",
                                "type": [
                                    "null",
                                    {
                                        "name": "C",
                                        "type": "record",
                                        "fields": [{"name": "foo", "type": "string"}],
                                    },
                                ],
                                "default": None,
                            }
                        ],
                    },
                ],
                "default": None,
            },
            {"name": "c", "type": ["null", "C"], "default": None},
        ],
    }
    assert loaded_schema == expected_schema


def test_load_schema_output_is_correct_4():
    """https://github.com/fastavro/fastavro/issues/476"""
    load_schema_dir = join(abspath(dirname(__file__)), "load_schema_test_4")
    schema_path = join(load_schema_dir, "A.avsc")
    loaded_schema = fastavro.schema.load_schema(schema_path, _write_hint=False)

    expected_schema = {
        "name": "A",
        "type": "record",
        "fields": [
            {
                "name": "b",
                "type": [
                    "null",
                    {
                        "name": "B",
                        "type": "record",
                        "fields": [{"name": "foo", "type": "string"}],
                    },
                ],
                "default": None,
            },
            {
                "name": "c",
                "type": [
                    "null",
                    {
                        "name": "C",
                        "type": "record",
                        "fields": [
                            {
                                "name": "b",
                                "type": ["null", "B"],
                                "default": None,
                            },
                        ],
                    },
                ],
                "default": None,
            },
        ],
    }
    assert loaded_schema == expected_schema


def test_load_schema_output_is_correct_5():
    """https://github.com/fastavro/fastavro/issues/476"""
    load_schema_dir = join(abspath(dirname(__file__)), "load_schema_test_5")
    schema_path = join(load_schema_dir, "A.avsc")
    loaded_schema = fastavro.schema.load_schema(schema_path, _write_hint=False)

    expected_schema = {
        "name": "A",
        "type": "record",
        "fields": [
            {
                "name": "map_field",
                "type": {
                    "type": "map",
                    "values": {
                        "name": "B",
                        "type": "record",
                        "fields": [{"name": "foo", "type": "string"}],
                    },
                },
            },
            {"name": "b", "type": "B"},
        ],
    }
    assert loaded_schema == expected_schema


def test_load_schema_output_is_correct_6():
    """https://github.com/fastavro/fastavro/issues/476"""
    load_schema_dir = join(abspath(dirname(__file__)), "load_schema_test_6")
    schema_path = join(load_schema_dir, "A.avsc")
    loaded_schema = fastavro.schema.load_schema(schema_path, _write_hint=False)

    expected_schema = {
        "name": "A",
        "type": "record",
        "fields": [
            {
                "name": "array_field",
                "type": {
                    "type": "array",
                    "items": {
                        "name": "B",
                        "type": "record",
                        "fields": [{"name": "foo", "type": "string"}],
                    },
                },
            },
            {"name": "b", "type": "B"},
        ],
    }
    assert loaded_schema == expected_schema


def test_load_schema_resolve_namespace():
    """https://github.com/fastavro/fastavro/issues/490"""
    load_schema_dir = join(abspath(dirname(__file__)), "load_schema_test_7")
    schema_path = join(load_schema_dir, "A.avsc")
    loaded_schema = fastavro.schema.load_schema(schema_path, _write_hint=False)

    expected_schema = {
        "name": "namespace.A",
        "type": "record",
        "fields": [
            {
                "name": "b",
                "type": {
                    "name": "namespace.B",
                    "type": "record",
                    "fields": [{"name": "foo", "type": "string"}],
                },
            },
        ],
    }
    assert loaded_schema == expected_schema


def test_load_schema_resolve_namespace_enums():
    """https://github.com/fastavro/fastavro/issues/490"""
    load_schema_dir = join(abspath(dirname(__file__)), "load_schema_test_8")
    schema_path = join(load_schema_dir, "A.avsc")
    loaded_schema = fastavro.schema.load_schema(schema_path, _write_hint=False)

    expected_schema = {
        "name": "namespace.A",
        "type": "record",
        "fields": [
            {
                "name": "b",
                "type": {
                    "name": "namespace.B",
                    "type": "enum",
                    "symbols": ["THIS", "THAT"],
                },
            },
            {
                "name": "c",
                "type": {
                    "name": "namespace.C",
                    "type": "enum",
                    "symbols": ["AND_THIS", "AND_THAT"],
                },
            },
        ],
    }
    assert loaded_schema == expected_schema


def test_load_schema_resolve_namespace_fixed():
    """https://github.com/fastavro/fastavro/issues/490"""
    load_schema_dir = join(abspath(dirname(__file__)), "load_schema_test_9")
    schema_path = join(load_schema_dir, "A.avsc")
    loaded_schema = fastavro.schema.load_schema(schema_path, _write_hint=False)

    expected_schema = {
        "name": "namespace.A",
        "type": "record",
        "fields": [
            {
                "name": "b",
                "type": {
                    "name": "namespace.B",
                    "type": "fixed",
                    "size": 4,
                },
            },
            {
                "name": "c",
                "type": {
                    "name": "namespace.C",
                    "type": "fixed",
                    "size": 8,
                },
            },
        ],
    }
    assert loaded_schema == expected_schema


def test_load_schema_bug():
    """https://github.com/fastavro/fastavro/issues/494"""
    load_schema_dir = join(abspath(dirname(__file__)), "load_schema_test_10")
    schema_path = join(load_schema_dir, "A.avsc")
    loaded_schema = fastavro.schema.load_schema(schema_path, _write_hint=False)

    expected_schema = {
        "name": "A",
        "type": "record",
        "fields": [
            {
                "name": "b",
                "type": [
                    "null",
                    {
                        "name": "B",
                        "type": "record",
                        "fields": [{"name": "foo", "type": "string"}],
                    },
                ],
                "default": None,
            },
            {
                "name": "c",
                "type": [
                    "null",
                    {
                        "name": "C",
                        "type": "record",
                        "fields": [
                            {
                                "name": "b",
                                "type": ["null", "B"],
                                "default": None,
                            },
                            {
                                "name": "d",
                                "type": [
                                    "null",
                                    {
                                        "name": "D",
                                        "type": "record",
                                        "fields": [{"name": "bar", "type": "string"}],
                                    },
                                ],
                                "default": None,
                            },
                        ],
                    },
                ],
                "default": None,
            },
        ],
    }
    assert loaded_schema == expected_schema


def test_load_schema_bug_2():
    """https://github.com/fastavro/fastavro/issues/494"""
    load_schema_dir = join(abspath(dirname(__file__)), "load_schema_test_11")
    schema_path = join(load_schema_dir, "A.avsc")
    loaded_schema = fastavro.schema.load_schema(schema_path, _write_hint=False)

    expected_schema = {
        "name": "A",
        "type": "record",
        "fields": [
            {
                "name": "b",
                "type": {
                    "name": "B",
                    "type": "record",
                    "fields": [
                        {
                            "name": "e",
                            "type": {
                                "name": "E",
                                "type": "record",
                                "fields": [{"name": "baz", "type": "string"}],
                            },
                        },
                    ],
                },
            },
            {
                "name": "c",
                "type": {
                    "name": "C",
                    "type": "record",
                    "fields": [
                        {
                            "name": "d",
                            "type": {
                                "name": "D",
                                "type": "record",
                                "fields": [{"name": "bar", "type": "string"}],
                            },
                        },
                        {"name": "e", "type": "E"},
                    ],
                },
            },
        ],
    }
    assert loaded_schema == expected_schema


def test_load_schema_ordered():
    """https://github.com/fastavro/fastavro/issues/493"""
    load_schema_dir = join(abspath(dirname(__file__)), "load_schema_test_12")
    load_order = [
        join(load_schema_dir, "com", "namespace", "E.avsc"),
        join(load_schema_dir, "com", "namespace", "D.avsc"),
        join(load_schema_dir, "com", "C.avsc"),
        join(load_schema_dir, "com", "B.avsc"),
        join(load_schema_dir, "A.avsc"),
    ]

    loaded_schema = fastavro.schema.load_schema_ordered(load_order, _write_hint=False)

    expected_schema = {
        "name": "A",
        "type": "record",
        "fields": [
            {
                "name": "b",
                "type": {
                    "name": "com.B",
                    "type": "record",
                    "fields": [
                        {
                            "name": "e",
                            "type": {
                                "name": "com.namespace.E",
                                "type": "record",
                                "fields": [{"name": "baz", "type": "string"}],
                            },
                        },
                    ],
                },
            },
            {
                "name": "c",
                "type": {
                    "name": "com.C",
                    "type": "record",
                    "fields": [
                        {
                            "name": "d",
                            "type": {
                                "name": "com.namespace.D",
                                "type": "record",
                                "fields": [{"name": "bar", "type": "string"}],
                            },
                        },
                        {"name": "e", "type": "com.namespace.E"},
                    ],
                },
            },
        ],
    }
    assert loaded_schema == expected_schema
